# Copyright 2015 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import errno

import mock
from six.moves import builtins

from os_win import constants
from os_win import exceptions
from os_win.tests.unit import test_base
from os_win.utils.io import namedpipe
from os_win.utils.winapi import constants as w_const


class NamedPipeTestCase(test_base.BaseTestCase):
    _FAKE_LOG_PATH = 'fake_log_path'

    @mock.patch.object(namedpipe.NamedPipeHandler, '_setup_io_structures')
    def setUp(self, mock_setup_structures):
        super(NamedPipeTestCase, self).setUp()

        self._mock_input_queue = mock.Mock()
        self._mock_output_queue = mock.Mock()
        self._mock_client_connected = mock.Mock()

        self._ioutils = mock.Mock()

        threading_patcher = mock.patch.object(namedpipe, 'threading')
        threading_patcher.start()
        self.addCleanup(threading_patcher.stop)

        self._handler = namedpipe.NamedPipeHandler(
            mock.sentinel.pipe_name,
            self._mock_input_queue,
            self._mock_output_queue,
            self._mock_client_connected,
            self._FAKE_LOG_PATH)
        self._handler._ioutils = self._ioutils

    def _mock_setup_pipe_handler(self):
        self._handler._log_file_handle = mock.Mock()
        self._handler._pipe_handle = mock.sentinel.pipe_handle
        self._r_worker = mock.Mock()
        self._w_worker = mock.Mock()
        self._handler._workers = [self._r_worker, self._w_worker]
        self._handler._r_buffer = mock.Mock()
        self._handler._w_buffer = mock.Mock()
        self._handler._r_overlapped = mock.Mock()
        self._handler._w_overlapped = mock.Mock()
        self._handler._r_completion_routine = mock.Mock()
        self._handler._w_completion_routine = mock.Mock()

    @mock.patch.object(builtins, 'open')
    @mock.patch.object(namedpipe.NamedPipeHandler, '_open_pipe')
    def test_start_pipe_handler(self, mock_open_pipe, mock_open):
        self._handler.start()

        mock_open_pipe.assert_called_once_with()
        mock_open.assert_called_once_with(self._FAKE_LOG_PATH, 'ab', 1)
        self.assertEqual(mock_open.return_value,
                         self._handler._log_file_handle)

        thread = namedpipe.threading.Thread
        thread.assert_has_calls(
            [mock.call(target=self._handler._read_from_pipe),
             mock.call().setDaemon(True),
             mock.call().start(),
             mock.call(target=self._handler._write_to_pipe),
             mock.call().setDaemon(True),
             mock.call().start()])

    @mock.patch.object(namedpipe.NamedPipeHandler, 'stop')
    @mock.patch.object(namedpipe.NamedPipeHandler, '_open_pipe')
    def test_start_pipe_handler_exception(self, mock_open_pipe,
                                          mock_stop_handler):
        mock_open_pipe.side_effect = Exception

        self.assertRaises(exceptions.OSWinException,
                          self._handler.start)

        mock_stop_handler.assert_called_once_with()

    @mock.patch.object(namedpipe.NamedPipeHandler, '_cleanup_handles')
    @mock.patch.object(namedpipe.NamedPipeHandler, '_cancel_io')
    def _test_stop_pipe_handler(self, mock_cancel_io,
                                mock_cleanup_handles,
                                workers_started=True):
        self._mock_setup_pipe_handler()
        if not workers_started:
            handler_workers = []
            self._handler._workers = handler_workers
        else:
            handler_workers = self._handler._workers
            self._r_worker.is_alive.side_effect = (True, False)
            self._w_worker.is_alive.return_value = False

        self._handler.stop()

        self._handler._stopped.set.assert_called_once_with()
        if not workers_started:
            mock_cleanup_handles.assert_called_once_with()
        else:
            self.assertFalse(mock_cleanup_handles.called)

        if workers_started:
            mock_cancel_io.assert_called_once_with()
            self._r_worker.join.assert_called_once_with(0.5)
            self.assertFalse(self._w_worker.join.called)

        self.assertEqual([], self._handler._workers)

    def test_stop_pipe_handler_workers_started(self):
        self._test_stop_pipe_handler()

    def test_stop_pipe_handler_workers_not_started(self):
        self._test_stop_pipe_handler(workers_started=False)

    @mock.patch.object(namedpipe.NamedPipeHandler, '_close_pipe')
    def test_cleanup_handles(self, mock_close_pipe):
        self._mock_setup_pipe_handler()
        log_handle = self._handler._log_file_handle
        r_event = self._handler._r_overlapped.hEvent
        w_event = self._handler._w_overlapped.hEvent

        self._handler._cleanup_handles()

        mock_close_pipe.assert_called_once_with()
        log_handle.close.assert_called_once_with()
        self._ioutils.close_handle.assert_has_calls(
            [mock.call(r_event), mock.call(w_event)])

        self.assertIsNone(self._handler._log_file_handle)
        self.assertIsNone(self._handler._r_overlapped.hEvent)
        self.assertIsNone(self._handler._w_overlapped.hEvent)

    def test_setup_io_structures(self):
        self._handler._setup_io_structures()

        self.assertEqual(self._ioutils.get_buffer.return_value,
                         self._handler._r_buffer)
        self.assertEqual(self._ioutils.get_buffer.return_value,
                         self._handler._w_buffer)
        self.assertEqual(
            self._ioutils.get_new_overlapped_structure.return_value,
            self._handler._r_overlapped)
        self.assertEqual(
            self._ioutils.get_new_overlapped_structure.return_value,
            self._handler._w_overlapped)
        self.assertEqual(
            self._ioutils.get_completion_routine.return_value,
            self._handler._r_completion_routine)
        self.assertEqual(
            self._ioutils.get_completion_routine.return_value,
            self._handler._w_completion_routine)
        self.assertIsNone(self._handler._log_file_handle)

        self._ioutils.get_buffer.assert_has_calls(
            [mock.call(constants.SERIAL_CONSOLE_BUFFER_SIZE)] * 2)
        self._ioutils.get_completion_routine.assert_has_calls(
            [mock.call(self._handler._read_callback),
             mock.call()])

    def test_open_pipe(self):
        self._handler._open_pipe()

        self._ioutils.wait_named_pipe.assert_called_once_with(
            mock.sentinel.pipe_name)
        self._ioutils.open.assert_called_once_with(
            mock.sentinel.pipe_name,
            desired_access=(w_const.GENERIC_READ | w_const.GENERIC_WRITE),
            share_mode=(w_const.FILE_SHARE_READ | w_const.FILE_SHARE_WRITE),
            creation_disposition=w_const.OPEN_EXISTING,
            flags_and_attributes=w_const.FILE_FLAG_OVERLAPPED)

        self.assertEqual(self._ioutils.open.return_value,
                         self._handler._pipe_handle)

    def test_close_pipe(self):
        self._mock_setup_pipe_handler()

        self._handler._close_pipe()

        self._ioutils.close_handle.assert_called_once_with(
            mock.sentinel.pipe_handle)
        self.assertIsNone(self._handler._pipe_handle)

    def test_cancel_io(self):
        self._mock_setup_pipe_handler()

        self._handler._cancel_io()

        overlapped_structures = [self._handler._r_overlapped,
                                 self._handler._w_overlapped]

        self._ioutils.cancel_io.assert_has_calls(
            [mock.call(self._handler._pipe_handle,
                       overlapped_structure,
                       ignore_invalid_handle=True)
             for overlapped_structure in overlapped_structures])

    @mock.patch.object(namedpipe.NamedPipeHandler, '_start_io_worker')
    def test_read_from_pipe(self, mock_start_worker):
        self._mock_setup_pipe_handler()

        self._handler._read_from_pipe()

        mock_start_worker.assert_called_once_with(
            self._ioutils.read,
            self._handler._r_buffer,
            self._handler._r_overlapped,
            self._handler._r_completion_routine)

    @mock.patch.object(namedpipe.NamedPipeHandler, '_start_io_worker')
    def test_write_to_pipe(self, mock_start_worker):
        self._mock_setup_pipe_handler()

        self._handler._write_to_pipe()

        mock_start_worker.assert_called_once_with(
            self._ioutils.write,
            self._handler._w_buffer,
            self._handler._w_overlapped,
            self._handler._w_completion_routine,
            self._handler._get_data_to_write)

    @mock.patch.object(namedpipe.NamedPipeHandler, '_cleanup_handles')
    def _test_start_io_worker(self, mock_cleanup_handles,
                              buff_update_func=None, exception=None):
        self._handler._stopped.isSet.side_effect = [False, True]
        self._handler._pipe_handle = mock.sentinel.pipe_handle
        self._handler.stop = mock.Mock()

        io_func = mock.Mock(side_effect=exception)
        fake_buffer = 'fake_buffer'

        self._handler._start_io_worker(io_func, fake_buffer,
                                       mock.sentinel.overlapped_structure,
                                       mock.sentinel.completion_routine,
                                       buff_update_func)

        if buff_update_func:
            num_bytes = buff_update_func()
        else:
            num_bytes = len(fake_buffer)

        io_func.assert_called_once_with(mock.sentinel.pipe_handle,
                                        fake_buffer, num_bytes,
                                        mock.sentinel.overlapped_structure,
                                        mock.sentinel.completion_routine)

        if exception:
            self._handler._stopped.set.assert_called_once_with()
        mock_cleanup_handles.assert_called_once_with()

    def test_start_io_worker(self):
        self._test_start_io_worker()

    def test_start_io_worker_with_buffer_update_method(self):
        self._test_start_io_worker(buff_update_func=mock.Mock())

    def test_start_io_worker_exception(self):
        self._test_start_io_worker(exception=IOError)

    @mock.patch.object(namedpipe.NamedPipeHandler, '_write_to_log')
    def test_read_callback(self, mock_write_to_log):
        self._mock_setup_pipe_handler()
        fake_data = self._ioutils.get_buffer_data.return_value

        self._handler._read_callback(mock.sentinel.num_bytes)

        self._ioutils.get_buffer_data.assert_called_once_with(
            self._handler._r_buffer, mock.sentinel.num_bytes)
        self._mock_output_queue.put.assert_called_once_with(fake_data)
        mock_write_to_log.assert_called_once_with(fake_data)

    @mock.patch.object(namedpipe, 'time')
    def test_get_data_to_write(self, mock_time):
        self._mock_setup_pipe_handler()
        self._handler._stopped.isSet.side_effect = [False, False]
        self._mock_client_connected.isSet.side_effect = [False, True]
        fake_data = 'fake input data'
        self._mock_input_queue.get.return_value = fake_data

        num_bytes = self._handler._get_data_to_write()

        mock_time.sleep.assert_called_once_with(1)
        self._ioutils.write_buffer_data.assert_called_once_with(
            self._handler._w_buffer, fake_data)
        self.assertEqual(len(fake_data), num_bytes)

    @mock.patch.object(namedpipe.NamedPipeHandler, '_rotate_logs')
    def _test_write_to_log(self, mock_rotate_logs, size_exceeded=False):
        self._mock_setup_pipe_handler()
        self._handler._stopped.isSet.return_value = False
        fake_handle = self._handler._log_file_handle
        fake_handle.tell.return_value = (constants.MAX_CONSOLE_LOG_FILE_SIZE
                                         if size_exceeded else 0)
        fake_data = 'fake_data'

        self._handler._write_to_log(fake_data)

        if size_exceeded:
            mock_rotate_logs.assert_called_once_with()

        self._handler._log_file_handle.write.assert_called_once_with(
            fake_data)

    def test_write_to_log(self):
        self._test_write_to_log()

    def test_write_to_log_size_exceeded(self):
        self._test_write_to_log(size_exceeded=True)

    def test_flush_log_file(self):
        self._handler._log_file_handle = None
        self._handler.flush_log_file()

        self._handler._log_file_handle = mock.Mock()
        self._handler.flush_log_file()

        self._handler._log_file_handle.flush.side_effect = ValueError
        self._handler.flush_log_file()

    @mock.patch.object(namedpipe.NamedPipeHandler, '_retry_if_file_in_use')
    @mock.patch.object(builtins, 'open')
    @mock.patch.object(namedpipe, 'os')
    def test_rotate_logs(self, mock_os, mock_open, mock_exec_retry):
        fake_archived_log_path = self._FAKE_LOG_PATH + '.1'
        mock_os.path.exists.return_value = True

        self._mock_setup_pipe_handler()
        fake_handle = self._handler._log_file_handle

        self._handler._rotate_logs()

        fake_handle.flush.assert_called_once_with()
        fake_handle.close.assert_called_once_with()
        mock_os.path.exists.assert_called_once_with(
            fake_archived_log_path)

        mock_exec_retry.assert_has_calls([mock.call(mock_os.remove,
                                                    fake_archived_log_path),
                                          mock.call(mock_os.rename,
                                                    self._FAKE_LOG_PATH,
                                                    fake_archived_log_path)])

        mock_open.assert_called_once_with(self._FAKE_LOG_PATH, 'ab', 1)
        self.assertEqual(mock_open.return_value,
                         self._handler._log_file_handle)

    @mock.patch.object(namedpipe, 'time')
    def test_retry_if_file_in_use_exceeded_retries(self, mock_time):
        class FakeWindowsException(Exception):
            errno = errno.EACCES

        raise_count = self._handler._MAX_LOG_ROTATE_RETRIES + 1
        mock_func_side_eff = [FakeWindowsException] * raise_count
        mock_func = mock.Mock(side_effect=mock_func_side_eff)

        with mock.patch.object(namedpipe, 'WindowsError',
                               FakeWindowsException, create=True):
            self.assertRaises(FakeWindowsException,
                              self._handler._retry_if_file_in_use,
                              mock_func, mock.sentinel.arg)
            mock_time.sleep.assert_has_calls(
                [mock.call(1)] * self._handler._MAX_LOG_ROTATE_RETRIES)



# Copyright 2015 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from os_win import constants
from os_win import exceptions
from os_win.tests.unit import test_base
from os_win.utils.storage.target import iscsi_target_utils as tg_utils


class ISCSITargetUtilsTestCase(test_base.OsWinBaseTestCase):

    _autospec_classes = [
        tg_utils.pathutils.PathUtils,
        tg_utils.hostutils.HostUtils,
        tg_utils.win32utils.Win32Utils,
    ]

    @mock.patch.object(tg_utils.ISCSITargetUtils,
                       '_ensure_wt_provider_available')
    def setUp(self, mock_ensure_wt_provider_available):
        super(ISCSITargetUtilsTestCase, self).setUp()

        self._tgutils = tg_utils.ISCSITargetUtils()
        self._tgutils._conn_wmi = mock.Mock()

    def test_ensure_wt_provider_unavailable(self):
        self._tgutils._conn_wmi = None
        self.assertRaises(exceptions.ISCSITargetException,
                          self._tgutils._ensure_wt_provider_available)

    def test_get_supported_disk_format_6_2(self):
        self._tgutils._win_gteq_6_3 = False
        fmt = self._tgutils.get_supported_disk_format()
        self.assertEqual(constants.DISK_FORMAT_VHD, fmt)

    def test_get_supported_disk_format_6_3(self):
        self._tgutils._win_gteq_6_3 = True
        fmt = self._tgutils.get_supported_disk_format()
        self.assertEqual(constants.DISK_FORMAT_VHDX, fmt)

    def test_get_supported_vhd_type_6_2(self):
        self._tgutils._win_gteq_6_3 = False
        vhd_type = self._tgutils.get_supported_vhd_type()
        self.assertEqual(constants.VHD_TYPE_FIXED, vhd_type)

    def test_get_supported_vhd_type_6_3(self):
        self._tgutils._win_gteq_6_3 = True
        vhd_type = self._tgutils.get_supported_vhd_type()
        self.assertEqual(constants.VHD_TYPE_DYNAMIC, vhd_type)

    def _test_get_portal_locations(self, available_only=False,
                                   fail_if_none_found=False):
        mock_portal = mock.Mock(Listen=False,
                                Address=mock.sentinel.address,
                                Port=mock.sentinel.port)
        mock_portal_location = "%s:%s" % (mock.sentinel.address,
                                          mock.sentinel.port)

        mock_wt_portal_cls = self._tgutils._conn_wmi.WT_Portal
        mock_wt_portal_cls.return_value = [mock_portal]

        if available_only and fail_if_none_found:
            self.assertRaises(exceptions.ISCSITargetException,
                              self._tgutils.get_portal_locations,
                              available_only=available_only,
                              fail_if_none_found=fail_if_none_found)
        else:
            portals = self._tgutils.get_portal_locations(
                available_only=available_only,
                fail_if_none_found=fail_if_none_found)

            expected_retrieved_portals = []
            if not available_only:
                expected_retrieved_portals.append(mock_portal_location)

            self.assertEqual(expected_retrieved_portals,
                             portals)

    def test_get_portal_locations(self):
        self._test_get_portal_locations()

    def test_get_available_portal_locations(self):
        self._test_get_portal_locations(available_only=True)

    def test_get_portal_locations_failing_if_none(self):
        self._test_get_portal_locations(available_only=True,
                                        fail_if_none_found=True)

    def _test_get_wt_host(self, host_found=True, fail_if_not_found=False):
        mock_wt_host = mock.Mock()
        mock_wt_host_cls = self._tgutils._conn_wmi.WT_Host
        mock_wt_host_cls.return_value = [mock_wt_host] if host_found else []

        if not host_found and fail_if_not_found:
            self.assertRaises(exceptions.ISCSITargetException,
                              self._tgutils._get_wt_host,
                              mock.sentinel.target_name,
                              fail_if_not_found=fail_if_not_found)
        else:
            wt_host = self._tgutils._get_wt_host(
                mock.sentinel.target_name,
                fail_if_not_found=fail_if_not_found)

            expected_wt_host = mock_wt_host if host_found else None
            self.assertEqual(expected_wt_host, wt_host)

        mock_wt_host_cls.assert_called_once_with(
            HostName=mock.sentinel.target_name)

    def test_get_wt_host(self):
        self._test_get_wt_host()

    def test_get_wt_host_not_found(self):
        self._test_get_wt_host(host_found=False)

    def test_get_wt_host_not_found_exception(self):
        self._test_get_wt_host(host_found=False,
                               fail_if_not_found=True)

    def _test_get_wt_disk(self, disk_found=True, fail_if_not_found=False):
        mock_wt_disk = mock.Mock()
        mock_wt_disk_cls = self._tgutils._conn_wmi.WT_Disk
        mock_wt_disk_cls.return_value = [mock_wt_disk] if disk_found else []

        if not disk_found and fail_if_not_found:
            self.assertRaises(exceptions.ISCSITargetException,
                              self._tgutils._get_wt_disk,
                              mock.sentinel.disk_description,
                              fail_if_not_found=fail_if_not_found)
        else:
            wt_disk = self._tgutils._get_wt_disk(
                mock.sentinel.disk_description,
                fail_if_not_found=fail_if_not_found)

            expected_wt_disk = mock_wt_disk if disk_found else None
            self.assertEqual(expected_wt_disk, wt_disk)

        mock_wt_disk_cls.assert_called_once_with(
            Description=mock.sentinel.disk_description)

    def test_get_wt_disk(self):
        self._test_get_wt_disk()

    def test_get_wt_disk_not_found(self):
        self._test_get_wt_disk(disk_found=False)

    def test_get_wt_disk_not_found_exception(self):
        self._test_get_wt_disk(disk_found=False,
                               fail_if_not_found=True)

    def _test_get_wt_snap(self, snap_found=True, fail_if_not_found=False):
        mock_wt_snap = mock.Mock()
        mock_wt_snap_cls = self._tgutils._conn_wmi.WT_Snapshot
        mock_wt_snap_cls.return_value = [mock_wt_snap] if snap_found else []

        if not snap_found and fail_if_not_found:
            self.assertRaises(exceptions.ISCSITargetException,
                              self._tgutils._get_wt_snapshot,
                              mock.sentinel.snap_description,
                              fail_if_not_found=fail_if_not_found)
        else:
            wt_snap = self._tgutils._get_wt_snapshot(
                mock.sentinel.snap_description,
                fail_if_not_found=fail_if_not_found)

            expected_wt_snap = mock_wt_snap if snap_found else None
            self.assertEqual(expected_wt_snap, wt_snap)

        mock_wt_snap_cls.assert_called_once_with(
            Description=mock.sentinel.snap_description)

    def test_get_wt_snap(self):
        self._test_get_wt_snap()

    def test_get_wt_snap_not_found(self):
        self._test_get_wt_snap(snap_found=False)

    def test_get_wt_snap_not_found_exception(self):
        self._test_get_wt_snap(snap_found=False,
                               fail_if_not_found=True)

    def _test_get_wt_idmethod(self, idmeth_found=True):
        mock_wt_idmeth = mock.Mock()
        mock_wt_idmeth_cls = self._tgutils._conn_wmi.WT_IDMethod
        mock_wt_idmeth_cls.return_value = ([mock_wt_idmeth]
                                           if idmeth_found else [])

        wt_idmeth = self._tgutils._get_wt_idmethod(mock.sentinel.initiator,
                                                   mock.sentinel.target_name)

        expected_wt_idmeth = mock_wt_idmeth if idmeth_found else None
        self.assertEqual(expected_wt_idmeth, wt_idmeth)

        mock_wt_idmeth_cls.assert_called_once_with(
            HostName=mock.sentinel.target_name,
            Value=mock.sentinel.initiator)

    def test_get_wt_idmethod(self):
        self._test_get_wt_idmethod()

    def test_get_wt_idmethod_not_found(self):
        self._test_get_wt_idmethod(idmeth_found=False)

    def _test_create_iscsi_target_exception(self, target_exists=False,
                                            fail_if_exists=False):
        mock_wt_host_cls = self._tgutils._conn_wmi.WT_Host
        mock_wt_host_cls.NewHost.side_effect = test_base.FakeWMIExc
        self._tgutils._win32utils.get_com_err_code.return_value = (
            self._tgutils._ERR_FILE_EXISTS if target_exists else 1)

        if target_exists and not fail_if_exists:
            self._tgutils.create_iscsi_target(mock.sentinel.target_name,
                                              fail_if_exists=fail_if_exists)
        else:
            self.assertRaises(exceptions.ISCSITargetException,
                              self._tgutils.create_iscsi_target,
                              mock.sentinel.target_name,
                              fail_if_exists=fail_if_exists)

        mock_wt_host_cls.NewHost.assert_called_once_with(
            HostName=mock.sentinel.target_name)

    def test_create_iscsi_target_exception(self):
        self._test_create_iscsi_target_exception()

    def test_create_iscsi_target_already_exists_skipping(self):
        self._test_create_iscsi_target_exception(target_exists=True)

    def test_create_iscsi_target_already_exists_failing(self):
        self._test_create_iscsi_target_exception(target_exists=True,
                                                 fail_if_exists=True)

    @mock.patch.object(tg_utils.ISCSITargetUtils, '_get_wt_host')
    def test_delete_iscsi_target_exception(self, mock_get_wt_host):
        mock_wt_host = mock_get_wt_host.return_value
        mock_wt_host.Delete_.side_effect = test_base.FakeWMIExc

        self.assertRaises(exceptions.ISCSITargetException,
                          self._tgutils.delete_iscsi_target,
                          mock.sentinel.target_name)

        mock_wt_host.RemoveAllWTDisks.assert_called_once_with()
        mock_get_wt_host.assert_called_once_with(mock.sentinel.target_name,
                                                 fail_if_not_found=False)

    @mock.patch.object(tg_utils.ISCSITargetUtils, '_get_wt_host')
    def _test_iscsi_target_exists(self, mock_get_wt_host, target_exists=True):
        mock_get_wt_host.return_value = (mock.sentinel.wt_host
                                         if target_exists else None)

        result = self._tgutils.iscsi_target_exists(mock.sentinel.target_name)

        self.assertEqual(target_exists, result)
        mock_get_wt_host.assert_called_once_with(mock.sentinel.target_name,
                                                 fail_if_not_found=False)

    def test_iscsi_target_exists(self):
        self._test_iscsi_target_exists()

    def test_iscsi_target_unexisting(self):
        self._test_iscsi_target_exists(target_exists=False)

    @mock.patch.object(tg_utils.ISCSITargetUtils, '_get_wt_host')
    def test_get_target_information(self, mock_get_wt_host):
        mock_wt_host = mock_get_wt_host.return_value
        mock_wt_host.EnableCHAP = True
        mock_wt_host.Status = 1  # connected

        target_info = self._tgutils.get_target_information(
            mock.sentinel.target_name)

        expected_info = dict(target_iqn=mock_wt_host.TargetIQN,
                             enabled=mock_wt_host.Enabled,
                             connected=True,
                             auth_method='CHAP',
                             auth_username=mock_wt_host.CHAPUserName,
                             auth_password=mock_wt_host.CHAPSecret)
        self.assertEqual(expected_info, target_info)
        mock_get_wt_host.assert_called_once_with(mock.sentinel.target_name)

    @mock.patch.object(tg_utils.ISCSITargetUtils, '_get_wt_host')
    def test_set_chap_credentials_exception(self, mock_get_wt_host):
        mock_wt_host = mock_get_wt_host.return_value
        mock_wt_host.put.side_effect = test_base.FakeWMIExc

        self.assertRaises(exceptions.ISCSITargetException,
                          self._tgutils.set_chap_credentials,
                          mock.sentinel.target_name,
                          mock.sentinel.chap_username,
                          mock.sentinel.chap_password)

        mock_get_wt_host.assert_called_once_with(mock.sentinel.target_name)
        self.assertTrue(mock_wt_host.EnableCHAP),
        self.assertEqual(mock.sentinel.chap_username,
                         mock_wt_host.CHAPUserName)
        self.assertEqual(mock.sentinel.chap_password,
                         mock_wt_host.CHAPSecret)
        mock_wt_host.put.assert_called_once_with()

    @mock.patch.object(tg_utils.ISCSITargetUtils, '_get_wt_idmethod')
    def test_associate_initiator_exception(self, mock_get_wtidmethod):
        mock_get_wtidmethod.return_value = None
        mock_wt_idmeth_cls = self._tgutils._conn_wmi.WT_IDMethod
        mock_wt_idmetod = mock_wt_idmeth_cls.new.return_value
        mock_wt_idmetod.put.side_effect = test_base.FakeWMIExc

        self.assertRaises(exceptions.ISCSITargetException,
                          self._tgutils.associate_initiator_with_iscsi_target,
                          mock.sentinel.initiator, mock.sentinel.target_name,
                          id_method=mock.sentinel.id_method)

        self.assertEqual(mock.sentinel.target_name, mock_wt_idmetod.HostName)
        self.assertEqual(mock.sentinel.initiator, mock_wt_idmetod.Value)
        self.assertEqual(mock.sentinel.id_method, mock_wt_idmetod.Method)
        mock_get_wtidmethod.assert_called_once_with(mock.sentinel.initiator,
                                                    mock.sentinel.target_name)

    @mock.patch.object(tg_utils.ISCSITargetUtils, '_get_wt_idmethod')
    def test_already_associated_initiator(self, mock_get_wtidmethod):
        mock_wt_idmeth_cls = self._tgutils._conn_wmi.WT_IDMethod

        self._tgutils.associate_initiator_with_iscsi_target(
            mock.sentinel.initiator, mock.sentinel.target_name,
            id_method=mock.sentinel.id_method)

        self.assertFalse(mock_wt_idmeth_cls.new.called)

    @mock.patch.object(tg_utils.ISCSITargetUtils, '_get_wt_idmethod')
    def test_deassociate_initiator_exception(self, mock_get_wtidmethod):
        mock_wt_idmetod = mock_get_wtidmethod.return_value
        mock_wt_idmetod.Delete_.side_effect = test_base.FakeWMIExc

        self.assertRaises(exceptions.ISCSITargetException,
                          self._tgutils.deassociate_initiator,
                          mock.sentinel.initiator, mock.sentinel.target_name)

        mock_get_wtidmethod.assert_called_once_with(mock.sentinel.initiator,
                                                    mock.sentinel.target_name)

    def test_create_wt_disk_exception(self):
        mock_wt_disk_cls = self._tgutils._conn_wmi.WT_Disk
        mock_wt_disk_cls.NewWTDisk.side_effect = test_base.FakeWMIExc

        self.assertRaises(exceptions.ISCSITargetException,
                          self._tgutils.create_wt_disk,
                          mock.sentinel.vhd_path, mock.sentinel.wtd_name,
                          mock.sentinel.size_mb)

        mock_wt_disk_cls.NewWTDisk.assert_called_once_with(
            DevicePath=mock.sentinel.vhd_path,
            Description=mock.sentinel.wtd_name,
            SizeInMB=mock.sentinel.size_mb)

    def test_import_wt_disk_exception(self):
        mock_wt_disk_cls = self._tgutils._conn_wmi.WT_Disk
        mock_wt_disk_cls.ImportWTDisk.side_effect = test_base.FakeWMIExc

        self.assertRaises(exceptions.ISCSITargetException,
                          self._tgutils.import_wt_disk,
                          mock.sentinel.vhd_path, mock.sentinel.wtd_name)

        mock_wt_disk_cls.ImportWTDisk.assert_called_once_with(
            DevicePath=mock.sentinel.vhd_path,
            Description=mock.sentinel.wtd_name)

    @mock.patch.object(tg_utils.ISCSITargetUtils, '_get_wt_disk')
    def test_change_wt_disk_status_exception(self, mock_get_wt_disk):
        mock_wt_disk = mock_get_wt_disk.return_value
        mock_wt_disk.put.side_effect = test_base.FakeWMIExc
        wt_disk_enabled = True

        self.assertRaises(exceptions.ISCSITargetException,
                          self._tgutils.change_wt_disk_status,
                          mock.sentinel.wtd_name,
                          enabled=wt_disk_enabled)

        mock_get_wt_disk.assert_called_once_with(mock.sentinel.wtd_name)
        self.assertEqual(wt_disk_enabled, mock_wt_disk.Enabled)

    @mock.patch.object(tg_utils.ISCSITargetUtils, '_get_wt_disk')
    def test_remove_wt_disk_exception(self, mock_get_wt_disk):
        mock_wt_disk = mock_get_wt_disk.return_value
        mock_wt_disk.Delete_.side_effect = test_base.FakeWMIExc

        self.assertRaises(exceptions.ISCSITargetException,
                          self._tgutils.remove_wt_disk,
                          mock.sentinel.wtd_name)

        mock_get_wt_disk.assert_called_once_with(mock.sentinel.wtd_name,
                                                 fail_if_not_found=False)

    @mock.patch.object(tg_utils.ISCSITargetUtils, '_get_wt_disk')
    def test_extend_wt_disk_exception(self, mock_get_wt_disk):
        mock_wt_disk = mock_get_wt_disk.return_value
        mock_wt_disk.Extend.side_effect = test_base.FakeWMIExc

        self.assertRaises(exceptions.ISCSITargetException,
                          self._tgutils.extend_wt_disk,
                          mock.sentinel.wtd_name,
                          mock.sentinel.additional_mb)

        mock_get_wt_disk.assert_called_once_with(mock.sentinel.wtd_name)
        mock_wt_disk.Extend.assert_called_once_with(
            mock.sentinel.additional_mb)

    @mock.patch.object(tg_utils.ISCSITargetUtils, '_get_wt_host')
    @mock.patch.object(tg_utils.ISCSITargetUtils, '_get_wt_disk')
    def test_add_disk_to_target_exception(self, mock_get_wt_disk,
                                          mock_get_wt_host):
        mock_wt_disk = mock_get_wt_disk.return_value
        mock_wt_host = mock_get_wt_host.return_value
        mock_wt_host.AddWTDisk.side_effect = test_base.FakeWMIExc

        self.assertRaises(exceptions.ISCSITargetException,
                          self._tgutils.add_disk_to_target,
                          mock.sentinel.wtd_name,
                          mock.sentinel.target_name)

        mock_get_wt_disk.assert_called_once_with(mock.sentinel.wtd_name)
        mock_get_wt_host.assert_called_once_with(mock.sentinel.target_name)
        mock_wt_host.AddWTDisk.assert_called_once_with(mock_wt_disk.WTD)

    @mock.patch.object(tg_utils.ISCSITargetUtils, '_get_wt_disk')
    def test_create_snapshot_exception(self, mock_get_wt_disk):
        mock_wt_disk = mock_get_wt_disk.return_value
        mock_wt_snap = mock.Mock()
        mock_wt_snap.put.side_effect = test_base.FakeWMIExc
        mock_wt_snap_cls = self._tgutils._conn_wmi.WT_Snapshot
        mock_wt_snap_cls.return_value = [mock_wt_snap]
        mock_wt_snap_cls.Create.return_value = [mock.sentinel.snap_id]

        self.assertRaises(exceptions.ISCSITargetException,
                          self._tgutils.create_snapshot,
                          mock.sentinel.wtd_name,
                          mock.sentinel.snap_name)

        mock_get_wt_disk.assert_called_once_with(mock.sentinel.wtd_name)
        mock_wt_snap_cls.Create.assert_called_once_with(WTD=mock_wt_disk.WTD)
        mock_wt_snap_cls.assert_called_once_with(Id=mock.sentinel.snap_id)
        self.assertEqual(mock.sentinel.snap_name, mock_wt_snap.Description)

    @mock.patch.object(tg_utils.ISCSITargetUtils, '_get_wt_snapshot')
    def test_delete_snapshot_exception(self, mock_get_wt_snap):
        mock_wt_snap = mock_get_wt_snap.return_value
        mock_wt_snap.Delete_.side_effect = test_base.FakeWMIExc

        self.assertRaises(exceptions.ISCSITargetException,
                          self._tgutils.delete_snapshot,
                          mock.sentinel.snap_name)

        mock_get_wt_snap.assert_called_once_with(mock.sentinel.snap_name,
                                                 fail_if_not_found=False)

    @mock.patch.object(tg_utils.ISCSITargetUtils, '_get_wt_snapshot')
    def test_export_snapshot_exception(self, mock_get_wt_snap):
        mock_wt_disk_cls = self._tgutils._conn_wmi.WT_Disk
        mock_wt_disk = mock.Mock()
        mock_wt_disk_cls.return_value = [mock_wt_disk]
        mock_wt_disk.Delete_.side_effect = test_base.FakeWMIExc
        mock_wt_snap = mock_get_wt_snap.return_value
        mock_wt_snap.Export.return_value = [mock.sentinel.wt_disk_id]

        self.assertRaises(exceptions.ISCSITargetException,
                          self._tgutils.export_snapshot,
                          mock.sentinel.snap_name,
                          mock.sentinel.dest_path)

        mock_get_wt_snap.assert_called_once_with(mock.sentinel.snap_name)
        mock_wt_snap.Export.assert_called_once_with()
        mock_wt_disk_cls.assert_called_once_with(WTD=mock.sentinel.wt_disk_id)

        expected_wt_disk_description = "%s-%s-temp" % (
            mock.sentinel.snap_name,
            mock.sentinel.wt_disk_id)
        self.assertEqual(expected_wt_disk_description,
                         mock_wt_disk.Description)

        mock_wt_disk.put.assert_called_once_with()
        mock_wt_disk.Delete_.assert_called_once_with()
        self._tgutils._pathutils.copy.assert_called_once_with(
            mock_wt_disk.DevicePath, mock.sentinel.dest_path)



# Copyright 2015 Cloudbase Solutions SRL
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit tests for the os_win._utils module.
"""

import ddt
import mock

from os_win import _utils
from os_win import constants
from os_win import exceptions
from os_win.tests.unit import test_base


@ddt.ddt
class UtilsTestCase(test_base.BaseTestCase):

    @mock.patch('oslo_concurrency.processutils.execute')
    def test_execute(self, mock_execute):
        _utils.execute(mock.sentinel.cmd, kwarg=mock.sentinel.kwarg)
        mock_execute.assert_called_once_with(mock.sentinel.cmd,
                                             kwarg=mock.sentinel.kwarg)

    def test_parse_server_string(self):
        result = _utils.parse_server_string('::1')
        self.assertEqual(('::1', ''), result)
        result = _utils.parse_server_string('[::1]:8773')
        self.assertEqual(('::1', '8773'), result)
        result = _utils.parse_server_string('2001:db8::192.168.1.1')
        self.assertEqual(('2001:db8::192.168.1.1', ''), result)
        result = _utils.parse_server_string('[2001:db8::192.168.1.1]:8773')
        self.assertEqual(('2001:db8::192.168.1.1', '8773'), result)
        result = _utils.parse_server_string('192.168.1.1')
        self.assertEqual(('192.168.1.1', ''), result)
        result = _utils.parse_server_string('192.168.1.2:8773')
        self.assertEqual(('192.168.1.2', '8773'), result)
        result = _utils.parse_server_string('192.168.1.3')
        self.assertEqual(('192.168.1.3', ''), result)
        result = _utils.parse_server_string('www.example.com:8443')
        self.assertEqual(('www.example.com', '8443'), result)
        result = _utils.parse_server_string('www.example.com')
        self.assertEqual(('www.example.com', ''), result)
        # error case
        result = _utils.parse_server_string('www.exa:mple.com:8443')
        self.assertEqual(('', ''), result)
        result = _utils.parse_server_string('')
        self.assertEqual(('', ''), result)

    def _get_fake_func_with_retry_decorator(self, side_effect,
                                            *args, **kwargs):
        func_side_effect = mock.Mock(side_effect=side_effect)

        @_utils.retry_decorator(*args, **kwargs)
        def fake_func(*_args, **_kwargs):
            return func_side_effect(*_args, **_kwargs)

        return fake_func, func_side_effect

    @mock.patch.object(_utils, 'time')
    def test_retry_decorator(self, mock_time):
        err_code = 1
        max_retry_count = 5
        max_sleep_time = 2
        timeout = max_retry_count + 1
        mock_time.time.side_effect = range(timeout)

        raised_exc = exceptions.Win32Exception(message='fake_exc',
                                               error_code=err_code)
        side_effect = [raised_exc] * max_retry_count
        side_effect.append(mock.sentinel.ret_val)

        (fake_func,
         fake_func_side_effect) = self._get_fake_func_with_retry_decorator(
            error_codes=err_code,
            exceptions=exceptions.Win32Exception,
            max_retry_count=max_retry_count,
            max_sleep_time=max_sleep_time,
            timeout=timeout,
            side_effect=side_effect)

        ret_val = fake_func(mock.sentinel.arg,
                            kwarg=mock.sentinel.kwarg)
        self.assertEqual(mock.sentinel.ret_val, ret_val)
        fake_func_side_effect.assert_has_calls(
            [mock.call(mock.sentinel.arg, kwarg=mock.sentinel.kwarg)] *
            (max_retry_count + 1))
        self.assertEqual(max_retry_count + 1, mock_time.time.call_count)
        mock_time.sleep.assert_has_calls(
            [mock.call(sleep_time)
             for sleep_time in [1, 2, 2, 2, 1]])

    @mock.patch.object(_utils, 'time')
    def _test_retry_decorator_exceeded(self, mock_time, expected_try_count,
                                       mock_time_side_eff=None,
                                       timeout=None, max_retry_count=None):
        raised_exc = exceptions.Win32Exception(message='fake_exc')
        mock_time.time.side_effect = mock_time_side_eff

        (fake_func,
         fake_func_side_effect) = self._get_fake_func_with_retry_decorator(
            exceptions=exceptions.Win32Exception,
            timeout=timeout,
            side_effect=raised_exc)

        self.assertRaises(exceptions.Win32Exception, fake_func)
        fake_func_side_effect.assert_has_calls(
            [mock.call()] * expected_try_count)

    def test_retry_decorator_tries_exceeded(self):
        self._test_retry_decorator_exceeded(
            max_retry_count=2,
            expected_try_count=3)

    def test_retry_decorator_time_exceeded(self):
        self._test_retry_decorator_exceeded(
            mock_time_side_eff=[0, 1, 4],
            timeout=3,
            expected_try_count=1)

    @mock.patch('time.sleep')
    def _test_retry_decorator_no_retry(self, mock_sleep,
                                       expected_exceptions=(),
                                       expected_error_codes=()):
        err_code = 1
        raised_exc = exceptions.Win32Exception(message='fake_exc',
                                               error_code=err_code)
        fake_func, fake_func_side_effect = (
            self._get_fake_func_with_retry_decorator(
                error_codes=expected_error_codes,
                exceptions=expected_exceptions,
                side_effect=raised_exc))

        self.assertRaises(exceptions.Win32Exception,
                          fake_func, mock.sentinel.arg,
                          fake_kwarg=mock.sentinel.kwarg)

        self.assertFalse(mock_sleep.called)
        fake_func_side_effect.assert_called_once_with(
            mock.sentinel.arg, fake_kwarg=mock.sentinel.kwarg)

    def test_retry_decorator_unexpected_err_code(self):
        self._test_retry_decorator_no_retry(
            expected_exceptions=exceptions.Win32Exception,
            expected_error_codes=2)

    def test_retry_decorator_unexpected_exc(self):
        self._test_retry_decorator_no_retry(
            expected_exceptions=(IOError, AttributeError))

    @mock.patch('time.sleep')
    def test_retry_decorator_explicitly_avoid_retry(self, mock_sleep):
        # Tests the case when there is a function aware of the retry
        # decorator and explicitly requests that no retry should be
        # performed.

        def func_side_effect(fake_arg, retry_context):
            self.assertEqual(mock.sentinel.arg, fake_arg)
            self.assertEqual(retry_context, dict(prevent_retry=False))

            retry_context['prevent_retry'] = True
            raise exceptions.Win32Exception(message='fake_exc',
                                            error_code=1)

        fake_func, mock_side_effect = (
            self._get_fake_func_with_retry_decorator(
                exceptions=exceptions.Win32Exception,
                side_effect=func_side_effect,
                pass_retry_context=True))

        self.assertRaises(exceptions.Win32Exception,
                          fake_func, mock.sentinel.arg)

        self.assertEqual(1, mock_side_effect.call_count)
        self.assertFalse(mock_sleep.called)

    @mock.patch.object(_utils.socket, 'getaddrinfo')
    def test_get_ips(self, mock_getaddrinfo):
        ips = ['1.2.3.4', '5.6.7.8']
        mock_getaddrinfo.return_value = [
            (None, None, None, None, (ip, 0)) for ip in ips]

        resulted_ips = _utils.get_ips(mock.sentinel.addr)
        self.assertEqual(ips, resulted_ips)

        mock_getaddrinfo.assert_called_once_with(
            mock.sentinel.addr, None, 0, 0, 0)

    @mock.patch('eventlet.tpool.execute')
    @mock.patch('eventlet.getcurrent')
    @ddt.data(mock.Mock(), None)
    def test_avoid_blocking_call(self, gt_parent, mock_get_current_gt,
                                 mock_execute):
        mock_get_current_gt.return_value.parent = gt_parent
        mock_execute.return_value = mock.sentinel.ret_val

        def fake_blocking_func(*args, **kwargs):
            self.assertEqual((mock.sentinel.arg, ), args)
            self.assertEqual(dict(kwarg=mock.sentinel.kwarg),
                             kwargs)
            return mock.sentinel.ret_val

        fake_blocking_func_decorated = (
            _utils.avoid_blocking_call_decorator(fake_blocking_func))

        ret_val = fake_blocking_func_decorated(mock.sentinel.arg,
                                               kwarg=mock.sentinel.kwarg)

        self.assertEqual(mock.sentinel.ret_val, ret_val)
        if gt_parent:
            mock_execute.assert_called_once_with(fake_blocking_func,
                                                 mock.sentinel.arg,
                                                 kwarg=mock.sentinel.kwarg)
        else:
            self.assertFalse(mock_execute.called)

    def test_get_com_error_hresult(self):
        fake_hres = -5
        expected_hres = (1 << 32) + fake_hres
        mock_excepinfo = [None] * 5 + [fake_hres]
        mock_com_err = mock.Mock(excepinfo=mock_excepinfo)

        ret_val = _utils.get_com_error_hresult(mock_com_err)

        self.assertEqual(expected_hres, ret_val)

    def get_com_error_hresult_missing_excepinfo(self):
        ret_val = _utils.get_com_error_hresult(None)
        self.assertIsNone(ret_val)

    @ddt.data(_utils._WBEM_E_NOT_FOUND, mock.sentinel.wbem_error)
    @mock.patch.object(_utils, 'get_com_error_hresult')
    def test_is_not_found_exc(self, hresult, mock_get_com_error_hresult):
        mock_get_com_error_hresult.return_value = hresult
        exc = mock.MagicMock()

        result = _utils._is_not_found_exc(exc)

        expected = hresult == _utils._WBEM_E_NOT_FOUND
        self.assertEqual(expected, result)
        mock_get_com_error_hresult.assert_called_once_with(exc.com_error)

    @mock.patch.object(_utils, 'get_com_error_hresult')
    def test_not_found_decorator(self, mock_get_com_error_hresult):
        mock_get_com_error_hresult.side_effect = lambda x: x
        translated_exc = exceptions.HyperVVMNotFoundException

        @_utils.not_found_decorator(
            translated_exc=translated_exc)
        def f(to_call):
            to_call()

        to_call = mock.Mock()
        to_call.side_effect = exceptions.x_wmi(
            'expected error', com_error=_utils._WBEM_E_NOT_FOUND)
        self.assertRaises(translated_exc, f, to_call)

        to_call.side_effect = exceptions.x_wmi()
        self.assertRaises(exceptions.x_wmi, f, to_call)

    def test_hex_str_to_byte_array(self):
        fake_hex_str = '0x0010A'

        resulted_array = _utils.hex_str_to_byte_array(fake_hex_str)
        expected_array = bytearray([0, 1, 10])

        self.assertEqual(expected_array, resulted_array)

    def test_byte_array_to_hex_str(self):
        fake_byte_array = bytearray(range(3))

        resulted_string = _utils.byte_array_to_hex_str(fake_byte_array)
        expected_string = '000102'

        self.assertEqual(expected_string, resulted_string)

    def test_required_vm_version(self):
        @_utils.required_vm_version()
        def foo(bar, vmsettings):
            pass

        mock_vmsettings = mock.Mock()

        for good_version in [constants.VM_VERSION_5_0,
                             constants.VM_VERSION_254_0]:
            mock_vmsettings.Version = good_version
            foo(mock.sentinel.bar, mock_vmsettings)

        for bad_version in ['4.99', '254.1']:
            mock_vmsettings.Version = bad_version
            self.assertRaises(exceptions.InvalidVMVersion, foo,
                              mock.sentinel.bar, mock_vmsettings)



# Copyright 2016 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import ctypes
import os
import re
import threading

from oslo_log import log as logging

from os_win._i18n import _
from os_win import _utils
from os_win import constants
from os_win import exceptions
from os_win.utils import baseutils
from os_win.utils import win32utils
from os_win.utils.winapi import libs as w_lib

kernel32 = w_lib.get_shared_lib_handle(w_lib.KERNEL32)

LOG = logging.getLogger(__name__)


class DEVICE_ID_VPD_PAGE(ctypes.BigEndianStructure):
    _fields_ = [
        ('DeviceType', ctypes.c_ubyte, 5),
        ('Qualifier', ctypes.c_ubyte, 3),
        ('PageCode', ctypes.c_ubyte),
        ('PageLength', ctypes.c_uint16)
    ]


class IDENTIFICATION_DESCRIPTOR(ctypes.Structure):
    _fields_ = [
        ('CodeSet', ctypes.c_ubyte, 4),
        ('ProtocolIdentifier', ctypes.c_ubyte, 4),
        ('IdentifierType', ctypes.c_ubyte, 4),
        ('Association', ctypes.c_ubyte, 2),
        ('_reserved', ctypes.c_ubyte, 1),
        ('Piv', ctypes.c_ubyte, 1),
        ('_reserved', ctypes.c_ubyte),
        ('IdentifierLength', ctypes.c_ubyte)
    ]


PDEVICE_ID_VPD_PAGE = ctypes.POINTER(DEVICE_ID_VPD_PAGE)
PIDENTIFICATION_DESCRIPTOR = ctypes.POINTER(IDENTIFICATION_DESCRIPTOR)

SCSI_ID_ASSOC_TYPE_DEVICE = 0
SCSI_ID_CODE_SET_BINARY = 1
SCSI_ID_CODE_SET_ASCII = 2

_RESCAN_LOCK = threading.Lock()


class DiskUtils(baseutils.BaseUtils):

    _wmi_cimv2_namespace = 'root/cimv2'
    _wmi_storage_namespace = 'root/microsoft/windows/storage'

    def __init__(self):
        self._conn_cimv2 = self._get_wmi_conn(self._wmi_cimv2_namespace)
        self._conn_storage = self._get_wmi_conn(self._wmi_storage_namespace)
        self._win32_utils = win32utils.Win32Utils()

        # Physical device names look like \\.\PHYSICALDRIVE1
        self._phys_dev_name_regex = re.compile(r'\\\\.*\\[a-zA-Z]*([\d]+)')

    def _get_disk_by_number(self, disk_number, msft_disk_cls=True):
        if msft_disk_cls:
            disk = self._conn_storage.Msft_Disk(Number=disk_number)
        else:
            disk = self._conn_cimv2.Win32_DiskDrive(Index=disk_number)

        if not disk:
            err_msg = _("Could not find the disk number %s")
            raise exceptions.DiskNotFound(err_msg % disk_number)
        return disk[0]

    def _get_disks_by_unique_id(self, unique_id, unique_id_format):
        # In some cases, multiple disks having the same unique id may be
        # exposed to the OS. This may happen if there are multiple paths
        # to the LUN and MPIO is not properly configured. This can be
        # valuable information to the caller.
        disks = self._conn_storage.Msft_Disk(UniqueId=unique_id,
                                             UniqueIdFormat=unique_id_format)
        if not disks:
            err_msg = _("Could not find any disk having unique id "
                        "'%(unique_id)s' and unique id format "
                        "'%(unique_id_format)s'")
            raise exceptions.DiskNotFound(err_msg % dict(
                unique_id=unique_id,
                unique_id_format=unique_id_format))
        return disks

    def get_disk_numbers_by_unique_id(self, unique_id, unique_id_format):
        disks = self._get_disks_by_unique_id(unique_id, unique_id_format)
        return [disk.Number for disk in disks]

    def get_disk_uid_and_uid_type(self, disk_number):
        disk = self._get_disk_by_number(disk_number)
        return disk.UniqueId, disk.UniqueIdFormat

    def is_mpio_disk(self, disk_number):
        disk = self._get_disk_by_number(disk_number)
        return disk.Path.lower().startswith(r'\\?\mpio')

    def refresh_disk(self, disk_number):
        disk = self._get_disk_by_number(disk_number)
        disk.Refresh()

    def get_device_name_by_device_number(self, device_number):
        disk = self._get_disk_by_number(device_number,
                                        msft_disk_cls=False)
        return disk.Name

    def get_device_number_from_device_name(self, device_name):
        matches = self._phys_dev_name_regex.findall(device_name)
        if matches:
            return matches[0]

        err_msg = _("Could not find device number for device: %s")
        raise exceptions.DiskNotFound(err_msg % device_name)

    def rescan_disks(self, merge_requests=False):
        """Perform a disk rescan.

        :param merge_requests: If this flag is set and a disk rescan is
                               already pending, we'll just wait for it to
                               finish without issuing a new rescan request.
        """
        if merge_requests:
            rescan_pending = _RESCAN_LOCK.locked()
            if rescan_pending:
                LOG.debug("A disk rescan is already pending. "
                          "Waiting for it to complete.")

            with _RESCAN_LOCK:
                if not rescan_pending:
                    self._rescan_disks()
        else:
            self._rescan_disks()

    @_utils.retry_decorator(exceptions=(exceptions.x_wmi,
                                        exceptions.OSWinException))
    def _rescan_disks(self):
        LOG.debug("Rescanning disks.")

        ret = self._conn_storage.Msft_StorageSetting.UpdateHostStorageCache()

        if isinstance(ret, collections.Iterable):
            ret = ret[0]

        if ret:
            err_msg = _("Rescanning disks failed. Error code: %s.")
            raise exceptions.OSWinException(err_msg % ret)

        LOG.debug("Finished rescanning disks.")

    def get_disk_capacity(self, path, ignore_errors=False):
        """Returns total/free space for a given directory."""
        norm_path = os.path.abspath(path)

        total_bytes = ctypes.c_ulonglong(0)
        free_bytes = ctypes.c_ulonglong(0)

        try:
            self._win32_utils.run_and_check_output(
                kernel32.GetDiskFreeSpaceExW,
                ctypes.c_wchar_p(norm_path),
                None,
                ctypes.pointer(total_bytes),
                ctypes.pointer(free_bytes),
                kernel32_lib_func=True)
            return total_bytes.value, free_bytes.value
        except exceptions.Win32Exception as exc:
            LOG.error("Could not get disk %(path)s capacity info. "
                      "Exception: %(exc)s",
                      dict(path=path,
                           exc=exc))
            if ignore_errors:
                return 0, 0
            else:
                raise exc

    def get_disk_size(self, disk_number):
        """Returns the disk size, given a physical disk number."""
        disk = self._get_disk_by_number(disk_number)
        return disk.Size

    def _parse_scsi_page_83(self, buff,
                            select_supported_identifiers=False):
        """Parse SCSI Device Identification VPD (page 0x83 data).

        :param buff: a byte array containing the SCSI page 0x83 data.
        :param select_supported_identifiers: select identifiers supported
            by Windows, in the order of precedence.
        :returns: a list of identifiers represented as dicts, containing
                  SCSI Unique IDs.
        """
        identifiers = []

        buff_sz = len(buff)
        buff = (ctypes.c_ubyte * buff_sz)(*bytearray(buff))

        vpd_pg_struct_sz = ctypes.sizeof(DEVICE_ID_VPD_PAGE)

        if buff_sz < vpd_pg_struct_sz:
            reason = _('Invalid VPD page data.')
            raise exceptions.SCSIPageParsingError(page='0x83',
                                                  reason=reason)

        vpd_page = ctypes.cast(buff, PDEVICE_ID_VPD_PAGE).contents
        vpd_page_addr = ctypes.addressof(vpd_page)
        total_page_sz = vpd_page.PageLength + vpd_pg_struct_sz

        if vpd_page.PageCode != 0x83:
            reason = _('Unexpected page code: %s') % vpd_page.PageCode
            raise exceptions.SCSIPageParsingError(page='0x83',
                                                  reason=reason)
        if total_page_sz > buff_sz:
            reason = _('VPD page overflow.')
            raise exceptions.SCSIPageParsingError(page='0x83',
                                                  reason=reason)
        if not vpd_page.PageLength:
            LOG.info('Page 0x83 data does not contain any '
                     'identification descriptors.')
            return identifiers

        id_desc_offset = vpd_pg_struct_sz
        while id_desc_offset < total_page_sz:
            id_desc_addr = vpd_page_addr + id_desc_offset
            # Remaining buffer size
            id_desc_buff_sz = buff_sz - id_desc_offset

            identifier = self._parse_scsi_id_desc(id_desc_addr,
                                                  id_desc_buff_sz)
            identifiers.append(identifier)

            id_desc_offset += identifier['raw_id_desc_size']

        if select_supported_identifiers:
            identifiers = self._select_supported_scsi_identifiers(identifiers)

        return identifiers

    def _parse_scsi_id_desc(self, id_desc_addr, buff_sz):
        """Parse SCSI VPD identification descriptor."""
        id_desc_struct_sz = ctypes.sizeof(IDENTIFICATION_DESCRIPTOR)

        if buff_sz < id_desc_struct_sz:
            reason = _('Identifier descriptor overflow.')
            raise exceptions.SCSIIdDescriptorParsingError(reason=reason)

        id_desc = IDENTIFICATION_DESCRIPTOR.from_address(id_desc_addr)
        id_desc_sz = id_desc_struct_sz + id_desc.IdentifierLength
        identifier_addr = id_desc_addr + id_desc_struct_sz

        if id_desc_sz > buff_sz:
            reason = _('Identifier overflow.')
            raise exceptions.SCSIIdDescriptorParsingError(reason=reason)

        identifier = (ctypes.c_ubyte *
                      id_desc.IdentifierLength).from_address(
                          identifier_addr)
        raw_id = bytearray(identifier)

        if id_desc.CodeSet == SCSI_ID_CODE_SET_ASCII:
            parsed_id = bytes(
                bytearray(identifier)).decode('ascii').strip('\x00')
        else:
            parsed_id = _utils.byte_array_to_hex_str(raw_id)

        id_dict = {
            'code_set': id_desc.CodeSet,
            'protocol': (id_desc.ProtocolIdentifier
                         if id_desc.Piv else None),
            'type': id_desc.IdentifierType,
            'association': id_desc.Association,
            'raw_id': raw_id,
            'id': parsed_id,
            'raw_id_desc_size': id_desc_sz,
        }
        return id_dict

    def _select_supported_scsi_identifiers(self, identifiers):
        # This method will filter out unsupported SCSI identifiers,
        # also sorting them based on the order of precedence.
        selected_identifiers = []

        for id_type in constants.SUPPORTED_SCSI_UID_FORMATS:
            for identifier in identifiers:
                if identifier['type'] == id_type:
                    selected_identifiers.append(identifier)

        return selected_identifiers

    def get_new_disk_policy(self):
        # This policy is also known as the 'SAN policy', describing
        # how new disks will be handled.
        storsetting = self._conn_storage.MSFT_StorageSetting.Get()[1]
        return storsetting.NewDiskPolicy

    def set_new_disk_policy(self, policy):
        """Sets the new disk policy, also known as SAN policy.

        :param policy: an integer value, one of the DISK_POLICY_*
                       values defined in os_win.constants.
        """
        self._conn_storage.MSFT_StorageSetting.Set(
            NewDiskPolicy=policy)

    def set_disk_online(self, disk_number):
        disk = self._get_disk_by_number(disk_number)
        err_code = disk.Online()[1]
        if err_code:
            err_msg = (_("Failed to bring disk '%(disk_number)s' online. "
                         "Error code: %(err_code)s.") %
                       dict(disk_number=disk_number,
                            err_code=err_code))
            raise exceptions.DiskUpdateError(message=err_msg)

    def set_disk_offline(self, disk_number):
        disk = self._get_disk_by_number(disk_number)
        err_code = disk.Offline()[1]
        if err_code:
            err_msg = (_("Failed to bring disk '%(disk_number)s' offline. "
                         "Error code: %(err_code)s.") %
                       dict(disk_number=disk_number,
                            err_code=err_code))
            raise exceptions.DiskUpdateError(message=err_msg)

    def set_disk_readonly_status(self, disk_number, read_only):
        disk = self._get_disk_by_number(disk_number)
        err_code = disk.SetAttributes(IsReadOnly=bool(read_only))[1]
        if err_code:
            err_msg = (_("Failed to set disk '%(disk_number)s' read-only "
                         "status to '%(read_only)s'. "
                         "Error code: %(err_code)s.") %
                       dict(disk_number=disk_number,
                            err_code=err_code,
                            read_only=bool(read_only)))
            raise exceptions.DiskUpdateError(message=err_msg)



# Copyright 2015 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ctypes

import mock
import six

from os_win import _utils
from os_win import exceptions
from os_win.tests.unit import test_base
from os_win.utils.storage.initiator import fc_utils
from os_win.utils.winapi.libs import hbaapi as fc_struct


class FCUtilsTestCase(test_base.BaseTestCase):
    """Unit tests for the Hyper-V FCUtils class."""

    _autospec_classes = [
        fc_utils.win32utils.Win32Utils,
        fc_utils.diskutils.DiskUtils,
    ]

    _FAKE_ADAPTER_NAME = 'fake_adapter_name'
    _FAKE_ADAPTER_WWN = list(range(8))

    def setUp(self):
        super(FCUtilsTestCase, self).setUp()
        self._setup_lib_mocks()

        self._fc_utils = fc_utils.FCUtils()
        self._diskutils = self._fc_utils._diskutils

        self._run_mocker = mock.patch.object(self._fc_utils,
                                             '_run_and_check_output')
        self._run_mocker.start()

        self._mock_run = self._fc_utils._run_and_check_output

        self.addCleanup(mock.patch.stopall)

    def _setup_lib_mocks(self):
        self._ctypes = mock.Mock()
        # This is used in order to easily make assertions on the variables
        # passed by reference.
        self._ctypes.byref = lambda x: (x, "byref")

        mock.patch.object(fc_utils, 'hbaapi', create=True).start()
        self._ctypes_mocker = mock.patch.object(fc_utils, 'ctypes',
                                                self._ctypes)
        self._ctypes_mocker.start()

    def test_run_and_check_output(self):
        self._run_mocker.stop()
        self._fc_utils._run_and_check_output(
            adapter_name=self._FAKE_ADAPTER_NAME)

        mock_win32_run = self._fc_utils._win32_utils.run_and_check_output
        mock_win32_run.assert_called_once_with(
            adapter_name=self._FAKE_ADAPTER_NAME,
            failure_exc=exceptions.FCWin32Exception)

    def test_get_wwn_struct_from_hex_str(self):
        wwn_b_array = list(range(8))
        wwn_str = _utils.byte_array_to_hex_str(wwn_b_array)

        wwn_struct = self._fc_utils._wwn_struct_from_hex_str(wwn_str)
        self.assertEqual(wwn_b_array, list(wwn_struct.wwn))

    def test_get_fc_hba_count(self):
        hba_count = self._fc_utils.get_fc_hba_count()

        fc_utils.hbaapi.HBA_GetNumberOfAdapters.assert_called_once_with()
        self.assertEqual(fc_utils.hbaapi.HBA_GetNumberOfAdapters.return_value,
                         hba_count)

    def test_open_adapter_by_name(self):
        self._ctypes_mocker.stop()

        self._mock_run.return_value = mock.sentinel.handle

        resulted_handle = self._fc_utils._open_adapter_by_name(
            self._FAKE_ADAPTER_NAME)

        args_list = self._mock_run.call_args_list[0][0]
        self.assertEqual(fc_utils.hbaapi.HBA_OpenAdapter, args_list[0])
        self.assertEqual(six.b(self._FAKE_ADAPTER_NAME), args_list[1].value)

        self.assertEqual(mock.sentinel.handle, resulted_handle)

    @mock.patch.object(fc_utils.fc_struct, 'HBA_HANDLE')
    def test_open_adapter_by_wwn(self, mock_hba_handle_struct):
        exp_handle = mock_hba_handle_struct.return_value
        resulted_handle = self._fc_utils._open_adapter_by_wwn(
            mock.sentinel.wwn)

        self.assertEqual(exp_handle, resulted_handle)

        self._mock_run.assert_called_once_with(
            fc_utils.hbaapi.HBA_OpenAdapterByWWN,
            self._ctypes.byref(exp_handle),
            mock.sentinel.wwn)

    def test_close_adapter(self):
        self._fc_utils._close_adapter(mock.sentinel.hba_handle)
        fc_utils.hbaapi.HBA_CloseAdapter.assert_called_once_with(
            mock.sentinel.hba_handle)

    @mock.patch.object(fc_utils.FCUtils, '_open_adapter_by_name')
    @mock.patch.object(fc_utils.FCUtils, '_close_adapter')
    def test_get_hba_handle_by_name(self, mock_close_adapter,
                                    mock_open_adapter):
        with self._fc_utils._get_hba_handle(
                adapter_name=self._FAKE_ADAPTER_NAME) as handle:
            self.assertEqual(mock_open_adapter.return_value, handle)
            mock_open_adapter.assert_called_once_with(
                self._FAKE_ADAPTER_NAME)
        mock_close_adapter.assert_called_once_with(
            mock_open_adapter.return_value)

    @mock.patch.object(fc_utils.FCUtils, '_open_adapter_by_wwn')
    @mock.patch.object(fc_utils.FCUtils, '_close_adapter')
    def test_get_hba_handle_by_wwn(self, mock_close_adapter,
                                   mock_open_adapter):
        with self._fc_utils._get_hba_handle(
                adapter_wwn_struct=mock.sentinel.wwn) as handle:
            self.assertEqual(mock_open_adapter.return_value, handle)
            mock_open_adapter.assert_called_once_with(mock.sentinel.wwn)
        mock_close_adapter.assert_called_once_with(
            mock_open_adapter.return_value)

    def test_get_hba_handle_missing_params(self):
        self.assertRaises(exceptions.FCException,
                          self._fc_utils._get_hba_handle().__enter__)

    def test_get_adapter_name(self):
        self._ctypes_mocker.stop()
        fake_adapter_index = 1

        def update_buff(func, adapter_index, buff):
            buff.value = six.b(self._FAKE_ADAPTER_NAME)

        self._mock_run.side_effect = update_buff

        resulted_adapter_name = self._fc_utils._get_adapter_name(
            fake_adapter_index)

        args_list = self._mock_run.call_args_list[0][0]

        self.assertEqual(fc_utils.hbaapi.HBA_GetAdapterName,
                         args_list[0])
        self.assertIsInstance(args_list[1], ctypes.c_uint32)
        self.assertEqual(fake_adapter_index, args_list[1].value)

        buff = ctypes.cast(args_list[2], ctypes.POINTER(
            ctypes.c_char * 256)).contents
        self.assertIsInstance(buff, ctypes.c_char * 256)
        self.assertEqual(self._FAKE_ADAPTER_NAME, resulted_adapter_name)

    @mock.patch.object(fc_struct, 'get_target_mapping_struct')
    def test_get_target_mapping(self, mock_get_target_mapping):
        fake_entry_count = 10
        hresults = [fc_utils.HBA_STATUS_ERROR_MORE_DATA,
                    fc_utils.HBA_STATUS_OK]
        mock_mapping = mock.Mock(NumberOfEntries=fake_entry_count)
        mock_get_target_mapping.return_value = mock_mapping
        self._mock_run.side_effect = hresults

        resulted_mapping = self._fc_utils._get_target_mapping(
            mock.sentinel.hba_handle)

        expected_calls = [
            mock.call(fc_utils.hbaapi.HBA_GetFcpTargetMapping,
                      mock.sentinel.hba_handle,
                      self._ctypes.byref(mock_mapping),
                      ignored_error_codes=[fc_utils.HBA_STATUS_ERROR_MORE_DATA]
                      )] * 2
        self._mock_run.assert_has_calls(expected_calls)
        self.assertEqual(mock_mapping, resulted_mapping)
        mock_get_target_mapping.assert_has_calls([mock.call(0),
                                                  mock.call(fake_entry_count)])

    @mock.patch.object(fc_struct, 'HBA_PortAttributes')
    def test_get_adapter_port_attributes(self, mock_class_HBA_PortAttributes):
        resulted_port_attributes = self._fc_utils._get_adapter_port_attributes(
            mock.sentinel.hba_handle, mock.sentinel.port_index)

        self._mock_run.assert_called_once_with(
            fc_utils.hbaapi.HBA_GetAdapterPortAttributes,
            mock.sentinel.hba_handle,
            mock.sentinel.port_index,
            self._ctypes.byref(mock_class_HBA_PortAttributes.return_value))

        self.assertEqual(mock_class_HBA_PortAttributes.return_value,
                         resulted_port_attributes)

    @mock.patch.object(fc_struct, 'HBA_AdapterAttributes')
    def test_get_adapter_attributes(self, mock_class_HBA_AdapterAttributes):
        resulted_hba_attributes = self._fc_utils._get_adapter_attributes(
            mock.sentinel.hba_handle)

        self._mock_run.assert_called_once_with(
            fc_utils.hbaapi.HBA_GetAdapterAttributes,
            mock.sentinel.hba_handle,
            self._ctypes.byref(mock_class_HBA_AdapterAttributes.return_value))

        self.assertEqual(mock_class_HBA_AdapterAttributes.return_value,
                         resulted_hba_attributes)

    @mock.patch.object(fc_utils.FCUtils, 'get_fc_hba_count')
    def test_get_fc_hba_ports_missing_hbas(self, mock_get_fc_hba_count):
        mock_get_fc_hba_count.return_value = 0

        resulted_hba_ports = self._fc_utils.get_fc_hba_ports()

        self.assertEqual([], resulted_hba_ports)

    @mock.patch.object(fc_utils.FCUtils, '_get_fc_hba_adapter_ports')
    @mock.patch.object(fc_utils.FCUtils, '_get_adapter_name')
    @mock.patch.object(fc_utils.FCUtils, 'get_fc_hba_count')
    def test_get_fc_hba_ports(self, mock_get_fc_hba_count,
                              mock_get_adapter_name,
                              mock_get_adapter_ports):
        fake_adapter_count = 3

        mock_get_adapter_name.side_effect = [Exception,
                                             mock.sentinel.adapter_name,
                                             mock.sentinel.adapter_name]
        mock_get_fc_hba_count.return_value = fake_adapter_count
        mock_get_adapter_ports.side_effect = [Exception,
                                              [mock.sentinel.port]]

        expected_hba_ports = [mock.sentinel.port]
        resulted_hba_ports = self._fc_utils.get_fc_hba_ports()
        self.assertEqual(expected_hba_ports, resulted_hba_ports)
        self.assertEqual(expected_hba_ports, resulted_hba_ports)

        mock_get_adapter_name.assert_has_calls(
            [mock.call(index) for index in range(fake_adapter_count)])
        mock_get_adapter_ports.assert_has_calls(
            [mock.call(mock.sentinel.adapter_name)] * 2)

    @mock.patch.object(fc_utils.FCUtils, '_open_adapter_by_name')
    @mock.patch.object(fc_utils.FCUtils, '_close_adapter')
    @mock.patch.object(fc_utils.FCUtils, '_get_adapter_port_attributes')
    @mock.patch.object(fc_utils.FCUtils, '_get_adapter_attributes')
    def test_get_fc_hba_adapter_ports(self, mock_get_adapter_attributes,
                                      mock_get_adapter_port_attributes,
                                      mock_close_adapter,
                                      mock_open_adapter):
        fake_port_count = 1
        fake_port_index = 0
        # Local WWNs
        fake_node_wwn = list(range(3))
        fake_port_wwn = list(range(3))

        mock_adapter_attributes = mock.MagicMock()
        mock_adapter_attributes.NumberOfPorts = fake_port_count
        mock_port_attributes = mock.MagicMock()
        mock_port_attributes.NodeWWN.wwn = fake_node_wwn
        mock_port_attributes.PortWWN.wwn = fake_port_wwn

        mock_get_adapter_attributes.return_value = mock_adapter_attributes
        mock_get_adapter_port_attributes.return_value = mock_port_attributes

        resulted_hba_ports = self._fc_utils._get_fc_hba_adapter_ports(
            mock.sentinel.adapter_name)

        expected_hba_ports = [{
            'node_name': _utils.byte_array_to_hex_str(fake_node_wwn),
            'port_name': _utils.byte_array_to_hex_str(fake_port_wwn)
        }]
        self.assertEqual(expected_hba_ports, resulted_hba_ports)

        mock_open_adapter.assert_called_once_with(mock.sentinel.adapter_name)
        mock_close_adapter.assert_called_once_with(
            mock_open_adapter(mock.sentinel.adapter_nam))
        mock_get_adapter_attributes.assert_called_once_with(
            mock_open_adapter.return_value)
        mock_get_adapter_port_attributes.assert_called_once_with(
            mock_open_adapter.return_value, fake_port_index)

    @mock.patch.object(fc_utils.FCUtils, '_wwn_struct_from_hex_str')
    @mock.patch.object(fc_utils.FCUtils, '_open_adapter_by_wwn')
    @mock.patch.object(fc_utils.FCUtils, '_close_adapter')
    @mock.patch.object(fc_utils.FCUtils, '_get_target_mapping')
    def test_get_fc_target_mapping(self, mock_get_target_mapping,
                                   mock_close_adapter, mock_open_adapter,
                                   mock_wwn_struct_from_hex_str):
        # Remote WWNs
        fake_node_wwn = list(range(8))
        fake_port_wwn = list(range(8)[::-1])

        mock_fcp_mappings = mock.MagicMock()
        mock_entry = mock.MagicMock()
        mock_entry.FcpId.NodeWWN.wwn = fake_node_wwn
        mock_entry.FcpId.PortWWN.wwn = fake_port_wwn
        mock_fcp_mappings.Entries = [mock_entry]
        mock_get_target_mapping.return_value = mock_fcp_mappings

        resulted_mappings = self._fc_utils.get_fc_target_mappings(
            mock.sentinel.local_wwnn)

        expected_mappings = [{
            'node_name': _utils.byte_array_to_hex_str(fake_node_wwn),
            'port_name': _utils.byte_array_to_hex_str(fake_port_wwn),
            'device_name': mock_entry.ScsiId.OSDeviceName,
            'lun': mock_entry.ScsiId.ScsiOSLun,
            'fcp_lun': mock_entry.FcpId.FcpLun
        }]
        self.assertEqual(expected_mappings, resulted_mappings)

        mock_wwn_struct_from_hex_str.assert_called_once_with(
            mock.sentinel.local_wwnn)
        mock_open_adapter.assert_called_once_with(
            mock_wwn_struct_from_hex_str.return_value)

        mock_close_adapter.assert_called_once_with(
            mock_open_adapter.return_value)

    def test_refresh_hba_configuration(self):
        self._fc_utils.refresh_hba_configuration()

        expected_func = fc_utils.hbaapi.HBA_RefreshAdapterConfiguration
        expected_func.assert_called_once_with()

    def test_send_scsi_inquiry_v2(self):
        self._ctypes_mocker.stop()

        fake_port_wwn = fc_struct.HBA_WWN()
        fake_remote_port_wwn = fc_struct.HBA_WWN()
        fake_fcp_lun = 11

        fake_cdb_byte_1 = 1
        fake_cdb_byte_2 = 0x80

        fake_resp = bytearray(range(200))
        fake_sense_data = bytearray(range(200)[::-1])
        fake_scsi_status = 5

        def mock_run(func, hba_handle, port_wwn_struct,
                     remote_port_wwn_struct, fcp_lun, cdb_byte1,
                     cdb_byte2, p_resp_buff, p_resp_buff_sz,
                     p_scsi_status, p_sense_buff, p_sense_buff_sz):
            self.assertEqual(fc_utils.hbaapi.HBA_ScsiInquiryV2, func)
            self.assertEqual(mock.sentinel.hba_handle, hba_handle)
            self.assertEqual(fake_port_wwn, port_wwn_struct)
            self.assertEqual(fake_remote_port_wwn, remote_port_wwn_struct)

            self.assertEqual(fake_fcp_lun, fcp_lun.value)
            self.assertEqual(fake_cdb_byte_1, cdb_byte1.value)
            self.assertEqual(fake_cdb_byte_2, cdb_byte2.value)

            resp_buff_sz = ctypes.cast(
                p_resp_buff_sz,
                ctypes.POINTER(ctypes.c_uint32)).contents
            sense_buff_sz = ctypes.cast(
                p_sense_buff_sz,
                ctypes.POINTER(ctypes.c_uint32)).contents
            scsi_status = ctypes.cast(
                p_scsi_status,
                ctypes.POINTER(ctypes.c_ubyte)).contents

            self.assertEqual(fc_utils.SCSI_INQ_BUFF_SZ, resp_buff_sz.value)
            self.assertEqual(fc_utils.SENSE_BUFF_SZ, sense_buff_sz.value)

            resp_buff_type = (ctypes.c_ubyte * resp_buff_sz.value)
            sense_buff_type = (ctypes.c_ubyte * sense_buff_sz.value)

            resp_buff = ctypes.cast(p_resp_buff,
                                    ctypes.POINTER(resp_buff_type)).contents
            sense_buff = ctypes.cast(p_sense_buff,
                                     ctypes.POINTER(sense_buff_type)).contents

            resp_buff[:len(fake_resp)] = fake_resp
            sense_buff[:len(fake_sense_data)] = fake_sense_data

            resp_buff_sz.value = len(fake_resp)
            sense_buff_sz.value = len(fake_sense_data)
            scsi_status.value = fake_scsi_status

        self._mock_run.side_effect = mock_run

        resp_buff = self._fc_utils._send_scsi_inquiry_v2(
            mock.sentinel.hba_handle,
            fake_port_wwn,
            fake_remote_port_wwn,
            fake_fcp_lun,
            fake_cdb_byte_1,
            fake_cdb_byte_2)

        self.assertEqual(fake_resp, bytearray(resp_buff[:len(fake_resp)]))

    @mock.patch.object(fc_utils.FCUtils, '_send_scsi_inquiry_v2')
    def test_get_scsi_device_id_vpd(self, mock_send_scsi_inq):
        self._fc_utils._get_scsi_device_id_vpd(
            mock.sentinel.hba_handle, mock.sentinel.port_wwn,
            mock.sentinel.remote_port_wwn, mock.sentinel.fcp_lun)

        mock_send_scsi_inq.assert_called_once_with(
            mock.sentinel.hba_handle, mock.sentinel.port_wwn,
            mock.sentinel.remote_port_wwn, mock.sentinel.fcp_lun,
            1, 0x83)

    @mock.patch.object(fc_utils.FCUtils, '_wwn_struct_from_hex_str')
    @mock.patch.object(fc_utils.FCUtils, '_open_adapter_by_wwn')
    @mock.patch.object(fc_utils.FCUtils, '_close_adapter')
    @mock.patch.object(fc_utils.FCUtils, '_get_scsi_device_id_vpd')
    def test_get_scsi_device_identifiers(self, mock_get_scsi_dev_id_vpd,
                                         mock_close_adapter, mock_open_adapter,
                                         mock_wwn_struct_from_hex_str):

        mock_wwn_struct_from_hex_str.side_effect = (
            mock.sentinel.local_wwnn_struct, mock.sentinel.local_wwpn_struct,
            mock.sentinel.remote_wwpn_struct)
        self._diskutils._parse_scsi_page_83.return_value = (
            mock.sentinel.identifiers)

        identifiers = self._fc_utils.get_scsi_device_identifiers(
            mock.sentinel.local_wwnn, mock.sentinel.local_wwpn,
            mock.sentinel.remote_wwpn, mock.sentinel.fcp_lun,
            mock.sentinel.select_supp_ids)

        self.assertEqual(mock.sentinel.identifiers, identifiers)

        mock_wwn_struct_from_hex_str.assert_has_calls(
            [mock.call(wwn)
             for wwn in (mock.sentinel.local_wwnn, mock.sentinel.local_wwpn,
                         mock.sentinel.remote_wwpn)])

        mock_get_scsi_dev_id_vpd.assert_called_once_with(
            mock_open_adapter.return_value,
            mock.sentinel.local_wwpn_struct,
            mock.sentinel.remote_wwpn_struct,
            mock.sentinel.fcp_lun)
        self._diskutils._parse_scsi_page_83.assert_called_once_with(
            mock_get_scsi_dev_id_vpd.return_value,
            select_supported_identifiers=mock.sentinel.select_supp_ids)

        mock_open_adapter.assert_called_once_with(
            mock.sentinel.local_wwnn_struct)
        mock_close_adapter.assert_called_once_with(
            mock_open_adapter.return_value)



# Copyright 2016 Cloudbase Solutions Srl
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import ctypes

import ddt
import mock
import six

from os_win import _utils
from os_win import constants
from os_win import exceptions
from os_win.tests.unit import test_base
from os_win.utils.storage.initiator import iscsi_utils
from os_win.utils.winapi import constants as w_const
from os_win.utils.winapi.errmsg import iscsierr
from os_win.utils.winapi.libs import iscsidsc as iscsi_struct


@ddt.ddt
class ISCSIInitiatorUtilsTestCase(test_base.OsWinBaseTestCase):
    """Unit tests for the Hyper-V ISCSIInitiatorUtils class."""

    _autospec_classes = [
        iscsi_utils.win32utils.Win32Utils,
        iscsi_utils.diskutils.DiskUtils,
    ]

    def setUp(self):
        super(ISCSIInitiatorUtilsTestCase, self).setUp()

        self._initiator = iscsi_utils.ISCSIInitiatorUtils()
        self._diskutils = self._initiator._diskutils

        self._iscsidsc = mock.patch.object(
            iscsi_utils, 'iscsidsc', create=True).start()

        self._run_mocker = mock.patch.object(self._initiator,
                                             '_run_and_check_output')
        self._mock_run = self._run_mocker.start()

        iscsi_utils.portal_map = collections.defaultdict(set)

    def _mock_ctypes(self):
        self._ctypes = mock.Mock()
        # This is used in order to easily make assertions on the variables
        # passed by reference.
        self._ctypes.byref = lambda x: (x, "byref")

        mock.patch.object(iscsi_utils, 'ctypes', self._ctypes).start()

    def _get_fake_iscsi_utils_getter_func(self, func_side_effect,
                                          decorator_args,
                                          returned_element_count=None,
                                          required_buff_sz=None):
        @iscsi_utils.ensure_buff_and_retrieve_items(**decorator_args)
        def fake_func(inst, buff=None, buff_size=None,
                      element_count=None, *args, **kwargs):
            raised_exc = None
            try:
                # Those arguments will always be ULONGs, as requested
                # by the iscsidsc functions.
                self.assertIsInstance(buff_size, ctypes.c_ulong)
                self.assertIsInstance(element_count, ctypes.c_ulong)
                func_side_effect(buff=buff, buff_size_val=buff_size.value,
                                 element_count_val=element_count.value,
                                 *args, **kwargs)
            except Exception as ex:
                raised_exc = ex

            if returned_element_count:
                element_count.value = returned_element_count
            if required_buff_sz:
                buff_size.value = required_buff_sz

            if raised_exc:
                raise raised_exc
            return mock.sentinel.ret_val
        return fake_func

    @mock.patch.object(iscsi_utils, '_get_items_from_buff')
    def _test_ensure_buff_decorator(self, mock_get_items,
                                    required_buff_sz=None,
                                    returned_element_count=None,
                                    parse_output=False):
        insufficient_buff_exc = exceptions.Win32Exception(
            message='fake_err_msg',
            error_code=w_const.ERROR_INSUFFICIENT_BUFFER)
        func_requests_buff_sz = required_buff_sz is not None
        struct_type = ctypes.c_uint

        decorator_args = dict(struct_type=struct_type,
                              parse_output=parse_output,
                              func_requests_buff_sz=func_requests_buff_sz)

        func_side_effect = mock.Mock(side_effect=(insufficient_buff_exc, None))
        fake_func = self._get_fake_iscsi_utils_getter_func(
            returned_element_count=returned_element_count,
            required_buff_sz=required_buff_sz,
            func_side_effect=func_side_effect,
            decorator_args=decorator_args)

        ret_val = fake_func(self._initiator, fake_arg=mock.sentinel.arg)
        if parse_output:
            self.assertEqual(mock_get_items.return_value, ret_val)
        else:
            self.assertEqual(mock.sentinel.ret_val, ret_val)

        # We expect our decorated method to be called exactly two times.
        first_call_args_dict = func_side_effect.call_args_list[0][1]
        self.assertIsInstance(first_call_args_dict['buff'],
                              ctypes.POINTER(struct_type))
        self.assertEqual(first_call_args_dict['buff_size_val'], 0)
        self.assertEqual(first_call_args_dict['element_count_val'], 0)

        second_call_args_dict = func_side_effect.call_args_list[1][1]
        self.assertIsInstance(second_call_args_dict['buff'],
                              ctypes.POINTER(struct_type))
        self.assertEqual(second_call_args_dict['buff_size_val'],
                         required_buff_sz or 0)
        self.assertEqual(second_call_args_dict['element_count_val'],
                         returned_element_count or 0)

    def test_ensure_buff_func_requests_buff_sz(self):
        self._test_ensure_buff_decorator(required_buff_sz=10,
                                         parse_output=True)

    def test_ensure_buff_func_requests_el_count(self):
        self._test_ensure_buff_decorator(returned_element_count=5)

    def test_ensure_buff_func_unexpected_exception(self):
        fake_exc = exceptions.Win32Exception(message='fake_message',
                                             error_code=1)

        func_side_effect = mock.Mock(side_effect=fake_exc)
        fake_func = self._get_fake_iscsi_utils_getter_func(
            func_side_effect=func_side_effect,
            decorator_args={'struct_type': ctypes.c_ubyte})

        self.assertRaises(exceptions.Win32Exception, fake_func,
                          self._initiator)

    def test_get_items_from_buff(self):
        fake_buff_contents = 'fake_buff_contents'
        fake_buff = (ctypes.c_wchar * len(fake_buff_contents))()
        fake_buff.value = fake_buff_contents

        fake_buff = ctypes.cast(fake_buff, ctypes.POINTER(ctypes.c_ubyte))

        result = iscsi_utils._get_items_from_buff(fake_buff, ctypes.c_wchar,
                                                  len(fake_buff_contents))

        self.assertEqual(fake_buff_contents, result.value)

    def test_run_and_check_output(self):
        self._run_mocker.stop()
        self._initiator._win32utils = mock.Mock()
        mock_win32utils_run_and_check_output = (
            self._initiator._win32utils.run_and_check_output)

        self._initiator._run_and_check_output(mock.sentinel.func,
                                              mock.sentinel.arg,
                                              fake_kwarg=mock.sentinel.kwarg)

        mock_win32utils_run_and_check_output.assert_called_once_with(
            mock.sentinel.func,
            mock.sentinel.arg,
            fake_kwarg=mock.sentinel.kwarg,
            error_msg_src=iscsierr.err_msg_dict,
            failure_exc=exceptions.ISCSIInitiatorAPIException)

    def test_get_iscsi_persistent_logins(self):
        self._mock_ctypes()

        _get_iscsi_persistent_logins = _utils.get_wrapped_function(
            self._initiator._get_iscsi_persistent_logins)
        _get_iscsi_persistent_logins(
            self._initiator,
            buff=mock.sentinel.buff,
            buff_size=mock.sentinel.buff_size,
            element_count=mock.sentinel.element_count)

        self._mock_run.assert_called_once_with(
            self._iscsidsc.ReportIScsiPersistentLoginsW,
            self._ctypes.byref(mock.sentinel.element_count),
            mock.sentinel.buff,
            self._ctypes.byref(mock.sentinel.buff_size))

    @mock.patch.object(iscsi_utils.ISCSIInitiatorUtils,
                       '_parse_string_list')
    def test_get_targets(self, mock_parse_string_list):
        self._mock_ctypes()

        get_targets = _utils.get_wrapped_function(
            self._initiator.get_targets)
        mock_el_count = mock.Mock(value=mock.sentinel.element_count)

        resulted_target_list = get_targets(
            self._initiator,
            forced_update=mock.sentinel.forced_update,
            element_count=mock_el_count,
            buff=mock.sentinel.buff)
        self.assertEqual(mock_parse_string_list.return_value,
                         resulted_target_list)

        self._mock_run.assert_called_once_with(
            self._iscsidsc.ReportIScsiTargetsW,
            mock.sentinel.forced_update,
            self._ctypes.byref(mock_el_count),
            mock.sentinel.buff)
        mock_parse_string_list.assert_called_once_with(
            mock.sentinel.buff, mock.sentinel.element_count)

    @mock.patch.object(iscsi_utils.ISCSIInitiatorUtils,
                       '_parse_string_list')
    def test_get_initiators(self, mock_parse_string_list):
        self._mock_ctypes()

        get_initiators = _utils.get_wrapped_function(
            self._initiator.get_iscsi_initiators)
        mock_el_count = mock.Mock(value=mock.sentinel.element_count)

        resulted_initator_list = get_initiators(
            self._initiator,
            element_count=mock_el_count,
            buff=mock.sentinel.buff)
        self.assertEqual(mock_parse_string_list.return_value,
                         resulted_initator_list)

        self._mock_run.assert_called_once_with(
            self._iscsidsc.ReportIScsiInitiatorListW,
            self._ctypes.byref(mock_el_count),
            mock.sentinel.buff)
        mock_parse_string_list.assert_called_once_with(
            mock.sentinel.buff, mock.sentinel.element_count)

    def test_parse_string_list(self):
        self._mock_ctypes()

        fake_buff = 'fake\x00buff\x00\x00'
        self._ctypes.cast.return_value = fake_buff

        str_list = self._initiator._parse_string_list(fake_buff,
                                                      len(fake_buff))

        self.assertEqual(['fake', 'buff'], str_list)

        self._ctypes.cast.assert_called_once_with(
            fake_buff, self._ctypes.POINTER.return_value)
        self._ctypes.POINTER.assert_called_once_with(self._ctypes.c_wchar)

    def test_get_iscsi_initiator(self):
        self._mock_ctypes()

        self._ctypes.c_wchar = mock.MagicMock()
        fake_buff = (self._ctypes.c_wchar * (
            w_const.MAX_ISCSI_NAME_LEN + 1))()
        fake_buff.value = mock.sentinel.buff_value

        resulted_iscsi_initiator = self._initiator.get_iscsi_initiator()

        self._mock_run.assert_called_once_with(
            self._iscsidsc.GetIScsiInitiatorNodeNameW,
            fake_buff)
        self.assertEqual(mock.sentinel.buff_value,
                         resulted_iscsi_initiator)

    @mock.patch('socket.getfqdn')
    def test_get_iscsi_initiator_exception(self, mock_get_fqdn):
        fake_fqdn = 'fakehost.FAKE-DOMAIN.com'
        fake_exc = exceptions.ISCSIInitiatorAPIException(
            message='fake_message',
            error_code=1,
            func_name='fake_func')

        self._mock_run.side_effect = fake_exc
        mock_get_fqdn.return_value = fake_fqdn

        resulted_iqn = self._initiator.get_iscsi_initiator()

        expected_iqn = "%s:%s" % (self._initiator._MS_IQN_PREFIX,
                                  fake_fqdn.lower())
        self.assertEqual(expected_iqn, resulted_iqn)

    @mock.patch.object(ctypes, 'byref')
    @mock.patch.object(iscsi_struct, 'ISCSI_UNIQUE_CONNECTION_ID')
    @mock.patch.object(iscsi_struct, 'ISCSI_UNIQUE_SESSION_ID')
    def test_login_iscsi_target(self, mock_cls_ISCSI_UNIQUE_SESSION_ID,
                                mock_cls_ISCSI_UNIQUE_CONNECTION_ID,
                                mock_byref):
        fake_target_name = 'fake_target_name'

        resulted_session_id, resulted_conection_id = (
            self._initiator._login_iscsi_target(fake_target_name))

        args_list = self._mock_run.call_args_list[0][0]

        self.assertIsInstance(args_list[1], ctypes.c_wchar_p)
        self.assertEqual(fake_target_name, args_list[1].value)
        self.assertIsInstance(args_list[4], ctypes.c_ulong)
        self.assertEqual(
            ctypes.c_ulong(w_const.ISCSI_ANY_INITIATOR_PORT).value,
            args_list[4].value)
        self.assertIsInstance(args_list[6], ctypes.c_ulonglong)
        self.assertEqual(0, args_list[6].value)
        self.assertIsInstance(args_list[9], ctypes.c_ulong)
        self.assertEqual(0, args_list[9].value)

        mock_byref.assert_has_calls([
            mock.call(mock_cls_ISCSI_UNIQUE_SESSION_ID.return_value),
            mock.call(mock_cls_ISCSI_UNIQUE_CONNECTION_ID.return_value)])
        self.assertEqual(
            mock_cls_ISCSI_UNIQUE_SESSION_ID.return_value,
            resulted_session_id)
        self.assertEqual(
            mock_cls_ISCSI_UNIQUE_CONNECTION_ID.return_value,
            resulted_conection_id)

    def test_get_iscsi_sessions(self):
        self._mock_ctypes()

        _get_iscsi_sessions = _utils.get_wrapped_function(
            self._initiator._get_iscsi_sessions)
        _get_iscsi_sessions(
            self._initiator,
            buff=mock.sentinel.buff,
            buff_size=mock.sentinel.buff_size,
            element_count=mock.sentinel.element_count)

        self._mock_run.assert_called_once_with(
            self._iscsidsc.GetIScsiSessionListW,
            self._ctypes.byref(mock.sentinel.buff_size),
            self._ctypes.byref(mock.sentinel.element_count),
            mock.sentinel.buff)

    @mock.patch.object(iscsi_utils.ISCSIInitiatorUtils,
                       '_get_iscsi_sessions')
    def test_get_iscsi_target_sessions(self, mock_get_iscsi_sessions,
                                       target_sessions_found=True):
        fake_session = mock.Mock(TargetNodeName="FAKE_TARGET_NAME",
                                 ConnectionCount=1)
        fake_disconn_session = mock.Mock(
            TargetNodeName="fake_target_name",
            ConnectionCount=0)
        other_session = mock.Mock(TargetNodeName="other_target_name",
                                  ConnectionCount=1)

        sessions = [fake_session, fake_disconn_session, other_session]
        mock_get_iscsi_sessions.return_value = sessions

        resulted_tgt_sessions = self._initiator._get_iscsi_target_sessions(
            "fake_target_name")

        self.assertEqual([fake_session], resulted_tgt_sessions)

    def test_get_iscsi_session_devices(self):
        self._mock_ctypes()

        _get_iscsi_session_devices = _utils.get_wrapped_function(
            self._initiator._get_iscsi_session_devices)
        _get_iscsi_session_devices(
            self._initiator,
            mock.sentinel.session_id,
            buff=mock.sentinel.buff,
            element_count=mock.sentinel.element_count)

        self._mock_run.assert_called_once_with(
            self._iscsidsc.GetDevicesForIScsiSessionW,
            self._ctypes.byref(mock.sentinel.session_id),
            self._ctypes.byref(mock.sentinel.element_count),
            mock.sentinel.buff)

    @mock.patch.object(iscsi_utils.ISCSIInitiatorUtils,
                       '_get_iscsi_session_devices')
    def test_get_iscsi_session_luns(self, mock_get_iscsi_session_devices):
        fake_device = mock.Mock()
        fake_device.StorageDeviceNumber.DeviceType = w_const.FILE_DEVICE_DISK
        mock_get_iscsi_session_devices.return_value = [fake_device,
                                                       mock.Mock()]

        resulted_luns = self._initiator._get_iscsi_session_disk_luns(
            mock.sentinel.session_id)
        expected_luns = [fake_device.ScsiAddress.Lun]

        mock_get_iscsi_session_devices.assert_called_once_with(
            mock.sentinel.session_id)
        self.assertEqual(expected_luns, resulted_luns)

    @mock.patch.object(iscsi_utils.ISCSIInitiatorUtils,
                       '_get_iscsi_session_devices')
    def test_get_iscsi_device_from_session(self,
                                           mock_get_iscsi_session_devices):
        fake_device = mock.Mock()
        fake_device.ScsiAddress.Lun = mock.sentinel.target_lun
        mock_get_iscsi_session_devices.return_value = [mock.Mock(),
                                                       fake_device]

        resulted_device = self._initiator._get_iscsi_device_from_session(
            mock.sentinel.session_id,
            mock.sentinel.target_lun)

        mock_get_iscsi_session_devices.assert_called_once_with(
            mock.sentinel.session_id)
        self.assertEqual(fake_device, resulted_device)

    @mock.patch.object(iscsi_utils.ISCSIInitiatorUtils,
                       'get_device_number_and_path')
    def test_get_device_number_for_target(self, mock_get_dev_num_and_path):
        dev_num = self._initiator.get_device_number_for_target(
            mock.sentinel.target_name, mock.sentinel.lun,
            mock.sentinel.fail_if_not_found)

        mock_get_dev_num_and_path.assert_called_once_with(
            mock.sentinel.target_name, mock.sentinel.lun,
            mock.sentinel.fail_if_not_found)
        self.assertEqual(mock_get_dev_num_and_path.return_value[0], dev_num)

    @mock.patch.object(iscsi_utils.ISCSIInitiatorUtils,
                       'ensure_lun_available')
    def test_get_device_number_and_path(self, mock_ensure_lun_available):
        mock_ensure_lun_available.return_value = (mock.sentinel.dev_num,
                                                  mock.sentinel.dev_path)

        dev_num, dev_path = self._initiator.get_device_number_and_path(
            mock.sentinel.target_name, mock.sentinel.lun,
            retry_attempts=mock.sentinel.retry_attempts,
            retry_interval=mock.sentinel.retry_interval,
            rescan_disks=mock.sentinel.rescan_disks,
            ensure_mpio_claimed=mock.sentinel.ensure_mpio_claimed)

        mock_ensure_lun_available.assert_called_once_with(
            mock.sentinel.target_name, mock.sentinel.lun,
            rescan_attempts=mock.sentinel.retry_attempts,
            retry_interval=mock.sentinel.retry_interval,
            rescan_disks=mock.sentinel.rescan_disks,
            ensure_mpio_claimed=mock.sentinel.ensure_mpio_claimed)

        self.assertEqual(mock.sentinel.dev_num, dev_num)
        self.assertEqual(mock.sentinel.dev_path, dev_path)

    @ddt.data(True, False)
    @mock.patch.object(iscsi_utils.ISCSIInitiatorUtils,
                       'ensure_lun_available')
    def test_get_device_number_and_path_exc(self, fail_if_not_found,
                                            mock_ensure_lun_available):
        raised_exc = exceptions.ISCSILunNotAvailable
        mock_ensure_lun_available.side_effect = raised_exc(
            target_iqn=mock.sentinel.target_iqn,
            target_lun=mock.sentinel.target_lun)

        if fail_if_not_found:
            self.assertRaises(raised_exc,
                              self._initiator.get_device_number_and_path,
                              mock.sentinel.target_name,
                              mock.sentinel.lun,
                              fail_if_not_found)
        else:
            dev_num, dev_path = self._initiator.get_device_number_and_path(
                mock.sentinel.target_name,
                mock.sentinel.lun,
                fail_if_not_found)
            self.assertIsNone(dev_num)
            self.assertIsNone(dev_path)

    @mock.patch.object(iscsi_utils.ISCSIInitiatorUtils,
                       '_get_iscsi_target_sessions')
    @mock.patch.object(iscsi_utils.ISCSIInitiatorUtils,
                       '_get_iscsi_session_disk_luns')
    def test_get_target_luns(self, mock_get_iscsi_session_disk_luns,
                             mock_get_iscsi_target_sessions):
        fake_session = mock.Mock()
        mock_get_iscsi_target_sessions.return_value = [fake_session]

        retrieved_luns = [mock.sentinel.lun_0]
        mock_get_iscsi_session_disk_luns.return_value = retrieved_luns

        resulted_luns = self._initiator.get_target_luns(
            mock.sentinel.target_name)

        mock_get_iscsi_target_sessions.assert_called_once_with(
            mock.sentinel.target_name)
        mock_get_iscsi_session_disk_luns.assert_called_once_with(
            fake_session.SessionId)
        self.assertEqual(retrieved_luns, resulted_luns)

    @mock.patch.object(iscsi_utils.ISCSIInitiatorUtils,
                       'get_target_luns')
    def test_get_target_lun_count(self, mock_get_target_luns):
        target_luns = [mock.sentinel.lun0, mock.sentinel.lun1]
        mock_get_target_luns.return_value = target_luns

        lun_count = self._initiator.get_target_lun_count(
            mock.sentinel.target_name)

        self.assertEqual(len(target_luns), lun_count)
        mock_get_target_luns.assert_called_once_with(
            mock.sentinel.target_name)

    def test_logout_iscsi_target(self):
        self._mock_ctypes()

        self._initiator._logout_iscsi_target(mock.sentinel.session_id)

        self._mock_run.assert_called_once_with(
            self._iscsidsc.LogoutIScsiTarget,
            self._ctypes.byref(mock.sentinel.session_id))

    def test_add_static_target(self):
        self._mock_ctypes()

        is_persistent = True
        self._initiator._add_static_target(mock.sentinel.target_name,
                                           is_persistent=is_persistent)

        self._mock_run.assert_called_once_with(
            self._iscsidsc.AddIScsiStaticTargetW,
            self._ctypes.c_wchar_p(mock.sentinel.target_name),
            None, 0, is_persistent, None, None, None)

    def test_remove_static_target(self):
        self._mock_ctypes()

        self._initiator._remove_static_target(mock.sentinel.target_name)

        expected_ignored_err_codes = [w_const.ISDSC_TARGET_NOT_FOUND]
        self._mock_run.assert_called_once_with(
            self._iscsidsc.RemoveIScsiStaticTargetW,
            self._ctypes.c_wchar_p(mock.sentinel.target_name),
            ignored_error_codes=expected_ignored_err_codes)

    def test_get_login_opts(self):
        fake_username = 'fake_chap_username'
        fake_password = 'fake_chap_secret'
        auth_type = constants.ISCSI_CHAP_AUTH_TYPE
        login_flags = w_const.ISCSI_LOGIN_FLAG_MULTIPATH_ENABLED

        login_opts = self._initiator._get_login_opts(
            auth_username=fake_username,
            auth_password=fake_password,
            auth_type=auth_type,
            login_flags=login_flags)

        self.assertEqual(len(fake_username), login_opts.UsernameLength)
        self.assertEqual(len(fake_password), login_opts.PasswordLength)

        username_struct_contents = ctypes.cast(
            login_opts.Username,
            ctypes.POINTER(ctypes.c_char * len(fake_username))).contents.value
        pwd_struct_contents = ctypes.cast(
            login_opts.Password,
            ctypes.POINTER(ctypes.c_char * len(fake_password))).contents.value

        self.assertEqual(six.b(fake_username), username_struct_contents)
        self.assertEqual(six.b(fake_password), pwd_struct_contents)

        expected_info_bitmap = (w_const.ISCSI_LOGIN_OPTIONS_USERNAME |
                                w_const.ISCSI_LOGIN_OPTIONS_PASSWORD |
                                w_const.ISCSI_LOGIN_OPTIONS_AUTH_TYPE)
        self.assertEqual(expected_info_bitmap,
                         login_opts.InformationSpecified)
        self.assertEqual(login_flags,
                         login_opts.LoginFlags)

    @mock.patch.object(iscsi_utils.ISCSIInitiatorUtils,
                       '_get_iscsi_session_devices')
    def test_session_on_path_exists(self, mock_get_iscsi_session_devices):
        mock_device = mock.Mock(InitiatorName=mock.sentinel.initiator_name)
        mock_get_iscsi_session_devices.return_value = [mock_device]

        fake_connection = mock.Mock(TargetAddress=mock.sentinel.portal_addr,
                                    TargetSocket=mock.sentinel.portal_port)
        fake_connections = [mock.Mock(), fake_connection]
        fake_session = mock.Mock(ConnectionCount=len(fake_connections),
                                 Connections=fake_connections)
        fake_sessions = [mock.Mock(Connections=[], ConnectionCount=0),
                         fake_session]

        session_on_path_exists = self._initiator._session_on_path_exists(
            fake_sessions, mock.sentinel.portal_addr,
            mock.sentinel.portal_port,
            mock.sentinel.initiator_name)
        self.assertTrue(session_on_path_exists)
        mock_get_iscsi_session_devices.assert_has_calls(
            [mock.call(session.SessionId) for session in fake_sessions])

    @mock.patch.object(iscsi_utils.ISCSIInitiatorUtils,
                       '_get_iscsi_target_sessions')
    @mock.patch.object(iscsi_utils.ISCSIInitiatorUtils,
                       '_session_on_path_exists')
    def _test_new_session_required(self, mock_session_on_path_exists,
                                   mock_get_iscsi_target_sessions,
                                   sessions=None,
                                   mpio_enabled=False,
                                   session_on_path_exists=False):
        mock_get_iscsi_target_sessions.return_value = sessions
        mock_session_on_path_exists.return_value = session_on_path_exists

        expected_result = (not sessions or
                           (mpio_enabled and not session_on_path_exists))
        result = self._initiator._new_session_required(
            mock.sentinel.target_iqn,
            mock.sentinel.portal_addr,
            mock.sentinel.portal_port,
            mock.sentinel.initiator_name,
            mpio_enabled)
        self.assertEqual(expected_result, result)

        if sessions and mpio_enabled:
            mock_session_on_path_exists.assert_called_once_with(
                sessions,
                mock.sentinel.portal_addr,
                mock.sentinel.portal_port,
                mock.sentinel.initiator_name)

    def test_new_session_required_no_sessions(self):
        self._test_new_session_required()

    def test_new_session_required_existing_sessions_no_mpio(self):
        self._test_new_session_required(sessions=mock.sentinel.sessions)

    def test_new_session_required_existing_sessions_mpio_enabled(self):
        self._test_new_session_required(sessions=mock.sentinel.sessions,
                                        mpio_enabled=True)

    def test_new_session_required_session_on_path_exists(self):
        self._test_new_session_required(sessions=mock.sentinel.sessions,
                                        mpio_enabled=True,
                                        session_on_path_exists=True)

    @mock.patch.object(iscsi_utils.ISCSIInitiatorUtils,
                       '_get_login_opts')
    @mock.patch.object(iscsi_struct, 'ISCSI_TARGET_PORTAL')
    @mock.patch.object(iscsi_utils.ISCSIInitiatorUtils,
                       '_new_session_required')
    @mock.patch.object(iscsi_utils.ISCSIInitiatorUtils, 'get_targets')
    @mock.patch.object(iscsi_utils.ISCSIInitiatorUtils, '_login_iscsi_target')
    @mock.patch.object(iscsi_utils.ISCSIInitiatorUtils,
                       'ensure_lun_available')
    @mock.patch.object(iscsi_utils.ISCSIInitiatorUtils,
                       '_add_static_target')
    def _test_login_storage_target(self, mock_add_static_target,
                                   mock_ensure_lun_available,
                                   mock_login_iscsi_target,
                                   mock_get_targets,
                                   mock_session_required,
                                   mock_cls_ISCSI_TARGET_PORTAL,
                                   mock_get_login_opts,
                                   mpio_enabled=False,
                                   login_required=True):
        fake_portal_addr = '127.0.0.1'
        fake_portal_port = 3260
        fake_target_portal = '%s:%s' % (fake_portal_addr, fake_portal_port)

        fake_portal = mock_cls_ISCSI_TARGET_PORTAL.return_value
        fake_login_opts = mock_get_login_opts.return_value

        mock_get_targets.return_value = []
        mock_login_iscsi_target.return_value = (mock.sentinel.session_id,
                                                mock.sentinel.conn_id)
        mock_session_required.return_value = login_required

        self._initiator.login_storage_target(
            mock.sentinel.target_lun,
            mock.sentinel.target_iqn,
            fake_target_portal,
            auth_username=mock.sentinel.auth_username,
            auth_password=mock.sentinel.auth_password,
            auth_type=mock.sentinel.auth_type,
            mpio_enabled=mpio_enabled,
            rescan_attempts=mock.sentinel.rescan_attempts)

        mock_get_targets.assert_called_once_with()
        mock_add_static_target.assert_called_once_with(
            mock.sentinel.target_iqn)

        if login_required:
            expected_login_flags = (
                w_const.ISCSI_LOGIN_FLAG_MULTIPATH_ENABLED
                if mpio_enabled else 0)
            mock_get_login_opts.assert_called_once_with(
                mock.sentinel.auth_username,
                mock.sentinel.auth_password,
                mock.sentinel.auth_type,
                expected_login_flags)
            mock_cls_ISCSI_TARGET_PORTAL.assert_called_once_with(
                Address=fake_portal_addr,
                Socket=fake_portal_port)
            mock_login_iscsi_target.assert_has_calls([
                mock.call(mock.sentinel.target_iqn,
                          fake_portal,
                          fake_login_opts,
                          is_persistent=True),
                mock.call(mock.sentinel.target_iqn,
                          fake_portal,
                          fake_login_opts,
                          is_persistent=False)])
        else:
            self.assertFalse(mock_login_iscsi_target.called)

        mock_ensure_lun_available.assert_called_once_with(
            mock.sentinel.target_iqn,
            mock.sentinel.target_lun,
            mock.sentinel.rescan_attempts)

    def test_login_storage_target_path_exists(self):
        self._test_login_storage_target(login_required=False)

    def test_login_new_storage_target_no_mpio(self):
        self._test_login_storage_target()

    def test_login_storage_target_new_path_using_mpio(self):
        self._test_login_storage_target(mpio_enabled=True)

    @ddt.data(dict(rescan_disks=True),
              dict(retry_interval=mock.sentinel.retry_interval))
    @ddt.unpack
    @mock.patch.object(iscsi_utils.ISCSIInitiatorUtils,
                       '_get_iscsi_device_from_session')
    @mock.patch.object(iscsi_utils.ISCSIInitiatorUtils,
                       '_get_iscsi_target_sessions')
    @mock.patch('time.sleep')
    def test_ensure_lun_available(self, mock_sleep,
                                  mock_get_iscsi_target_sessions,
                                  mock_get_iscsi_device_from_session,
                                  rescan_disks=False, retry_interval=0):
        retry_count = 6
        mock_get_iscsi_target_sessions.return_value = [
            mock.Mock(SessionId=mock.sentinel.session_id)]

        fake_exc = exceptions.ISCSIInitiatorAPIException(
            message='fake_message',
            error_code=1,
            func_name='fake_func')
        dev_num_side_eff = [None, -1] + [mock.sentinel.dev_num] * 3
        dev_path_side_eff = ([mock.sentinel.dev_path] * 2 +
                             [None] + [mock.sentinel.dev_path] * 2)
        fake_device = mock.Mock()
        type(fake_device.StorageDeviceNumber).DeviceNumber = (
            mock.PropertyMock(side_effect=dev_num_side_eff))
        type(fake_device).LegacyName = (
            mock.PropertyMock(side_effect=dev_path_side_eff))

        mock_get_dev_side_eff = [None, fake_exc] + [fake_device] * 5
        mock_get_iscsi_device_from_session.side_effect = mock_get_dev_side_eff
        self._diskutils.is_mpio_disk.side_effect = [False, True]

        dev_num, dev_path = self._initiator.ensure_lun_available(
            mock.sentinel.target_iqn,
            mock.sentinel.target_lun,
            rescan_attempts=retry_count,
            retry_interval=retry_interval,
            rescan_disks=rescan_disks,
            ensure_mpio_claimed=True)

        self.assertEqual(mock.sentinel.dev_num, dev_num)
        self.assertEqual(mock.sentinel.dev_path, dev_path)

        mock_get_iscsi_target_sessions.assert_has_calls(
            [mock.call(mock.sentinel.target_iqn)] * (retry_count + 1))
        mock_get_iscsi_device_from_session.assert_has_calls(
            [mock.call(mock.sentinel.session_id,
                       mock.sentinel.target_lun)] * retry_count)
        self._diskutils.is_mpio_disk.assert_has_calls(
            [mock.call(mock.sentinel.dev_num)] * 2)

        expected_rescan_count = retry_count if rescan_disks else 0
        self.assertEqual(
            expected_rescan_count,
            self._diskutils.rescan_disks.call_count)

        if retry_interval:
            mock_sleep.assert_has_calls(
                [mock.call(retry_interval)] * retry_count)
        else:
            self.assertFalse(mock_sleep.called)

    @mock.patch.object(iscsi_utils.ISCSIInitiatorUtils,
                       '_get_iscsi_target_sessions')
    @mock.patch.object(iscsi_utils.ISCSIInitiatorUtils,
                       '_logout_iscsi_target')
    @mock.patch.object(iscsi_utils.ISCSIInitiatorUtils,
                       '_remove_target_persistent_logins')
    @mock.patch.object(iscsi_utils.ISCSIInitiatorUtils,
                       '_remove_static_target')
    def test_logout_storage_target(self, mock_remove_static_target,
                                   mock_remove_target_persistent_logins,
                                   mock_logout_iscsi_target,
                                   mock_get_iscsi_target_sessions):
        fake_session = mock.Mock(SessionId=mock.sentinel.session_id)
        mock_get_iscsi_target_sessions.return_value = [fake_session]

        self._initiator.logout_storage_target(mock.sentinel.target_iqn)

        mock_get_iscsi_target_sessions.assert_called_once_with(
            mock.sentinel.target_iqn, connected_only=False)
        mock_logout_iscsi_target.assert_called_once_with(
            mock.sentinel.session_id)
        mock_remove_target_persistent_logins.assert_called_once_with(
            mock.sentinel.target_iqn)
        mock_remove_static_target.assert_called_once_with(
            mock.sentinel.target_iqn)

    @mock.patch.object(iscsi_utils.ISCSIInitiatorUtils,
                       '_remove_persistent_login')
    @mock.patch.object(iscsi_utils.ISCSIInitiatorUtils,
                       '_get_iscsi_persistent_logins')
    def test_remove_target_persistent_logins(self,
                                             mock_get_iscsi_persistent_logins,
                                             mock_remove_persistent_login):
        fake_persistent_login = mock.Mock(TargetName=mock.sentinel.target_iqn)
        mock_get_iscsi_persistent_logins.return_value = [fake_persistent_login]

        self._initiator._remove_target_persistent_logins(
            mock.sentinel.target_iqn)

        mock_remove_persistent_login.assert_called_once_with(
            fake_persistent_login)
        mock_get_iscsi_persistent_logins.assert_called_once_with()

    @mock.patch.object(ctypes, 'byref')
    def test_remove_persistent_login(self, mock_byref):
        fake_persistent_login = mock.Mock()
        fake_persistent_login.InitiatorInstance = 'fake_initiator_instance'
        fake_persistent_login.TargetName = 'fake_target_name'

        self._initiator._remove_persistent_login(fake_persistent_login)

        args_list = self._mock_run.call_args_list[0][0]
        self.assertIsInstance(args_list[1], ctypes.c_wchar_p)
        self.assertEqual(fake_persistent_login.InitiatorInstance,
                         args_list[1].value)
        self.assertIsInstance(args_list[3], ctypes.c_wchar_p)
        self.assertEqual(fake_persistent_login.TargetName,
                         args_list[3].value)
        mock_byref.assert_called_once_with(fake_persistent_login.TargetPortal)



#  Copyright 2016 Cloudbase Solutions Srl
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import six

from os_win.tests.unit import test_base
from os_win.utils import baseutils


class BaseUtilsTestCase(test_base.OsWinBaseTestCase):
    """Unit tests for the os-win BaseUtils class."""

    def setUp(self):
        super(BaseUtilsTestCase, self).setUp()
        self.utils = baseutils.BaseUtils()
        self.utils._conn = mock.MagicMock()

    @mock.patch.object(baseutils, 'wmi', create=True)
    def test_get_wmi_obj(self, mock_wmi):
        result = self.utils._get_wmi_obj(mock.sentinel.moniker)

        self.assertEqual(mock_wmi.WMI.return_value, result)
        mock_wmi.WMI.assert_called_once_with(moniker=mock.sentinel.moniker)

    @mock.patch.object(baseutils.BaseUtils, '_get_wmi_obj')
    @mock.patch.object(baseutils, 'sys')
    def _check_get_wmi_conn(self, mock_sys, mock_get_wmi_obj, **kwargs):
        mock_sys.platform = 'win32'
        result = self.utils._get_wmi_conn(mock.sentinel.moniker, **kwargs)

        self.assertEqual(mock_get_wmi_obj.return_value, result)
        mock_get_wmi_obj.assert_called_once_with(mock.sentinel.moniker,
                                                 **kwargs)

    def test_get_wmi_conn_kwargs(self):
        self.utils._WMI_CONS.clear()
        self._check_get_wmi_conn(privileges=mock.sentinel.privileges)
        self.assertNotIn(mock.sentinel.moniker, baseutils.BaseUtils._WMI_CONS)

    def test_get_wmi_conn(self):
        self._check_get_wmi_conn()
        self.assertIn(mock.sentinel.moniker, baseutils.BaseUtils._WMI_CONS)

    @mock.patch.object(baseutils.BaseUtils, '_get_wmi_obj')
    @mock.patch.object(baseutils, 'sys')
    def test_get_wmi_conn_cached(self, mock_sys, mock_get_wmi_obj):
        mock_sys.platform = 'win32'
        baseutils.BaseUtils._WMI_CONS[mock.sentinel.moniker] = (
            mock.sentinel.conn)
        result = self.utils._get_wmi_conn(mock.sentinel.moniker)

        self.assertEqual(mock.sentinel.conn, result)
        self.assertFalse(mock_get_wmi_obj.called)

    @mock.patch.object(baseutils, 'sys')
    def test_get_wmi_conn_linux(self, mock_sys):
        mock_sys.platform = 'linux'
        result = self.utils._get_wmi_conn(mock.sentinel.moniker)

        self.assertIsNone(result)


class BaseUtilsVirtTestCase(test_base.OsWinBaseTestCase):
    """Unit tests for the os-win BaseUtilsVirt class."""

    def setUp(self):
        super(BaseUtilsVirtTestCase, self).setUp()
        self.utils = baseutils.BaseUtilsVirt()
        self.utils._conn_attr = mock.MagicMock()
        baseutils.BaseUtilsVirt._os_version = None

    @mock.patch.object(baseutils.BaseUtilsVirt, '_get_wmi_conn')
    def test_conn(self, mock_get_wmi_conn):
        self.utils._conn_attr = None

        self.assertEqual(mock_get_wmi_conn.return_value, self.utils._conn)
        mock_get_wmi_conn.assert_called_once_with(
            self.utils._wmi_namespace % '.')

    def test_vs_man_svc(self):
        mock_os = mock.MagicMock(Version='6.3.0')
        self._mock_wmi.WMI.return_value.Win32_OperatingSystem.return_value = [
            mock_os]
        expected = self.utils._conn.Msvm_VirtualSystemManagementService()[0]
        self.assertEqual(expected, self.utils._vs_man_svc)
        self.assertEqual(expected, self.utils._vs_man_svc_attr)

    @mock.patch.object(baseutils, 'imp')
    @mock.patch.object(baseutils, 'wmi', create=True)
    def test_vs_man_svc_2012(self, mock_wmi, mock_imp):
        baseutils.BaseUtilsVirt._old_wmi = None
        mock_os = mock.MagicMock(Version='6.2.0')
        mock_wmi.WMI.return_value.Win32_OperatingSystem.return_value = [
            mock_os]
        fake_module_path = '/fake/path/to/module'
        mock_wmi.__path__ = [fake_module_path]
        old_conn = mock_imp.load_source.return_value.WMI.return_value

        expected = old_conn.Msvm_VirtualSystemManagementService()[0]
        self.assertEqual(expected, self.utils._vs_man_svc)
        self.assertIsNone(self.utils._vs_man_svc_attr)
        mock_imp.load_source.assert_called_once_with(
            'old_wmi', '%s.py' % fake_module_path)

    @mock.patch.object(baseutils.BaseUtilsVirt, '_get_wmi_compat_conn')
    def test_get_wmi_obj_compatibility_6_3(self, mock_get_wmi_compat):
        mock_os = mock.MagicMock(Version='6.3.0')
        self._mock_wmi.WMI.return_value.Win32_OperatingSystem.return_value = [
            mock_os]

        result = self.utils._get_wmi_obj(mock.sentinel.moniker, True)
        self.assertEqual(self._mock_wmi.WMI.return_value, result)

    @mock.patch.object(baseutils.BaseUtilsVirt, '_get_wmi_compat_conn')
    def test_get_wmi_obj_no_compatibility_6_2(self, mock_get_wmi_compat):
        baseutils.BaseUtilsVirt._os_version = [6, 2]
        result = self.utils._get_wmi_obj(mock.sentinel.moniker, False)
        self.assertEqual(self._mock_wmi.WMI.return_value, result)

    @mock.patch.object(baseutils.BaseUtilsVirt, '_get_wmi_compat_conn')
    def test_get_wmi_obj_compatibility_6_2(self, mock_get_wmi_compat):
        baseutils.BaseUtilsVirt._os_version = [6, 2]
        result = self.utils._get_wmi_obj(mock.sentinel.moniker, True)
        self.assertEqual(mock_get_wmi_compat.return_value, result)


class SynchronizedMetaTestCase(test_base.OsWinBaseTestCase):
    @mock.patch.object(baseutils.threading, 'RLock')
    def test_synchronized_meta(self, mock_rlock_cls):
        fake_cls = type('fake_cls', (object, ),
                        dict(method1=lambda x: None, method2=lambda y: None))
        fake_cls = six.add_metaclass(baseutils.SynchronizedMeta)(fake_cls)

        fake_cls().method1()
        fake_cls().method2()

        mock_rlock_cls.assert_called_once_with()
        self.assertEqual(2, mock_rlock_cls.return_value.__exit__.call_count)
